# coding=utf-8
"""
Purpose:   [1] To disambiguate the azimuthal component of the SDO/HMI vector magnetic field data (See Section 5 of Hoeksema et al. 2014 [open-access at https://link.springer.com/article/10.1007%2Fs11207-014-0516-8])
           [2] To decompose the three components of the vector magnetic field -- field strength, disambiguated azimuth, and inclination -- and project them into one of two coordinate systems: CCD or CEA (See Sun, X. 2013 [https://arxiv.org/abs/1309.2392] and Bobra et al. 2014 [open-access at https://link.springer.com/article/10.1007%2Fs11207-014-0529-3]).

Usage:     This code depends on the NumPy, SciPy, AstroPy, SunPy, and drms libraries.
           The first three libraries are in the standard Anaconda distribution.
           SunPy can be installed via conda: http://docs.sunpy.org/en/stable/guide/installation/
           The drms library can be obtained from pip: http://drms.readthedocs.io/en/stable/intro.html#installation.
           This code is compatible with python 3.5.x.

Examples:  See example disambiguate_data.py in the same repository.

Adapted:   From Xudong Sun's IDL code to do the same thing

RAW: https://github.com/mbobra/SHARPs/blob/master/disambiguation.py
"""

# import some modules
import sys
import numpy as np
import astropy.units as u
# import pandas
from astropy.io import fits
from astropy.coordinates import SkyCoord
import sunpy.coordinates
from sunpy.map import Map
from datetime import datetime as dt_obj
import drms

__all__ = ['Basic', 'CoordinateTransform']
__author__ = 'Monica Bobra'


class Basic(object):
    """
    Class for the basic disambiguation functions.

    Attributes
    ----------
    recordset: string
        Single recordset specification, e.g. 'hmi.sharp_720s[377][2011.02.15_00:00:00]'

    method: int
        Method used to disambiguate the data: 0 for potential acute, 1 for random, 2 for radial acute (suggested)
    """

    def __init__(self, recordset, method):
        self.recordset = recordset
        self.method = method

    # ===========================================
    @staticmethod
    def get_data(self):
        """function: get_data
        This function reads the appropriate data and metadata.

        Returns
        -------
        result: list
            List containing five items:
            [0] The relevant WCS keywords as a pandas dataframe
            [1] FITS file containing azimuthal component of magnetic field vector as an astropy HDUList object
            [2] FITS file containing field strength of magnetic field vector as an astropy HDUList object
            [3] FITS file containing inclination component of magnetic field vector as an astropy HDUList object
            [4] FITS file containing disambiguation information as an astropy HDUList object
        """

        c = drms.Client()

        try:
            keys, segments = c.query(
                self.recordset, key=drms.const.all, seg='inclination, azimuth, field, disambig')
            # print(type(keys))
            # print(keys)
            # keys.to_csv("keys.csv")
            # aaa = pandas.read_csv("keys.csv")
            # print(aaa)
        except:
            print("Invalid recordset specification")
            sys.exit(1)

        if (len(keys) > 1):
            print("Specify only one record")
            sys.exit(1)

        baseurl = 'http://jsoc.stanford.edu'
        # print("download from:\n {0}\n{1}\n{2}\n{3}\n".format(
        #     baseurl + segments.azimuth[0],
        #     baseurl + segments.field[0],
        #     baseurl + segments.inclination[0],
        #     baseurl + segments.disambig[0]))
        azimuth = fits.open(baseurl + segments.azimuth[0])
        field = fits.open(baseurl + segments.field[0])
        inclination = fits.open(baseurl + segments.inclination[0])
        disambig = fits.open(baseurl + segments.disambig[0])

        return [keys, azimuth, field, inclination, disambig]

    # ===========================================
    @staticmethod
    def perform_disambiguation(self, azimuth, disambig):
        """function: perform_disambiguation
        This function performs the actual disambiguation.

        Parameters
        ----------
        azimuth: astropy HDUList object
            FITS file containing azimuthal component of magnetic field vector

        disambig: astropy HDUList object
            FITS file containing disambiguation information

        Returns
        -------
        result: astropy HDUList object
            FITS file containing disambiguated azimuthal component of magnetic field vector
        """

        size_azimuth = azimuth[1].shape
        size_disambig = disambig[1].shape
        if (size_disambig != size_azimuth):
            print("file_azimuth and file_disambig are not of the same dimensions")
            sys.exit(1)

        if (self.method < 0 or self.method > 2):
            method = 2
            print('Invalid disambiguation method, set to default method = 2')

        disambig[1].data = (disambig[1].data /
                            (np.power(2, self.method))).astype('uint8')
        disambig[1].data = disambig[1].data * 180.
        azimuth[1].data = disambig[1].data + azimuth[1].data
        print("Disambiguated the data.")

        return azimuth


class CoordinateTransform(object):
    """
    Class for coordinate transformations.

    Attributes
    ----------
    azimuth: astropy HDUList object
        FITS file containing disambiguated azimuthal component of magnetic field vector

    field: astropy HDUList object
        FITS file containing field strength of magnetic field vector

    inclination: astropy HDUList object
        FITS file containing inclination component of magnetic field vector

    keys: pandas dataframe
        The relevant WCS keywords
    """

    def __init__(self, azimuth, field, inclination, keys):
        self.azimuth = azimuth
        self.field = field
        self.inclination = inclination
        self.keys = keys

    # ===========================================
    @staticmethod
    def ccd(self):
        """function: ccd
        This function constructs HMI vector field from its native coordinates (field, inclination, and disambiguated azimuth)
        to spherical coordinate components (B_phi, meridional B_theta, radial B_r) on the CCD grid.

        Returns
        -------
        latlon: astropy SkyCoord object (see: http://docs.astropy.org/en/stable/api/astropy.coordinates.SkyCoord.html)
            Astropy coordinate frame in heliographic stonyhurst coordinates

        bptr: numpy.ndarray
            Array containing the three spherical components of the vector magnetic field in
            heliographic stonyhurst coordinates (Bp, Bt, Br) on the CCD grid
        """

        size_field = self.field[1].shape
        size_inclination = self.inclination[1].shape
        size_azimuth = self.azimuth[1].shape
        if (size_field != size_inclination != size_azimuth):
            print("The three components are not of the same dimensions")
            print("Field", size_field, "inclination",
                  size_inclination, "azimuth", size_azimuth)
            sys.exit(1)

            # convert to radians
        gamma = self.inclination[1].data * (np.pi / 180.)
        psi = self.azimuth[1].data * (np.pi / 180.)

        # convert to b_xi, b_eta, and b_zeta as defined in Equation 1 of Sun et al. 2013
        b_xi = -1 * self.field[1].data * np.sin(gamma) * np.sin(psi)
        b_eta = self.field[1].data * np.sin(gamma) * np.cos(psi)
        b_zeta = self.field[1].data * np.cos(gamma)

        # --- convert from helioprojective cartesian to stonyhurst heliographic coordinates ---

        # step 1: convert from pixels to helioprojective cartesian coordinates
        coord_x = np.ndarray(
            [self.azimuth[1].header['NAXIS2'], self.azimuth[1].header['NAXIS1']])
        coord_y = np.ndarray(
            [self.azimuth[1].header['NAXIS2'], self.azimuth[1].header['NAXIS1']])

        # data[1].header gives the uncompressed header
        for pix_x in range(self.azimuth[1].header['NAXIS2']):
            for pix_y in range(self.azimuth[1].header['NAXIS1']):
                coord_x[pix_x, pix_y] = pix_x + 1
                coord_y[pix_x, pix_y] = pix_y + 1

        x1 = (((coord_x - self.keys.CRPIX1[0]) * np.cos(self.keys.CROTA2[0] * (np.pi / 180.)) - (
                coord_y - self.keys.CRPIX2[0]) * np.sin(self.keys.CROTA2[0] * (np.pi / 180.))) * self.keys.CDELT1[
                  0] + self.keys.CRVAL1[0])
        y1 = (((coord_y - self.keys.CRPIX2[0]) * np.cos(self.keys.CROTA2[0] * (np.pi / 180.)) + (
                coord_x - self.keys.CRPIX1[0]) * np.sin(self.keys.CROTA2[0] * (np.pi / 180.))) * self.keys.CDELT1[
                  0] + self.keys.CRVAL2[0])

        # step 2: populate these values into an astropy coordinate frame

        # the input dateobs takes a string! convert t_rec to a string of the required format
        def parse_tai_string(tstr, datetime=True):
            year = int(tstr[:4])
            month = int(tstr[5:7])
            day = int(tstr[8:10])
            hour = int(tstr[11:13])
            minute = int(tstr[14:16])
            if datetime:
                return dt_obj(year, month, day, hour, minute)
            else:
                return year, month, day, hour, minute

        out = parse_tai_string(self.keys.T_REC[0])
        dateobs_out = out.strftime('%Y/%m/%dT%H:%M:%S')

        hmimap = sunpy.map.Map(self.field[1].data, dict(self.keys.iloc[0]))
        c = SkyCoord(x1 * u.arcsec, y1 * u.arcsec, frame='helioprojective', rsun=self.keys.RSUN_REF[0] * u.meter,
                     observer=hmimap.observer_coordinate)

        # step 3: transform the skycoord frame from helioprojective cartesian to stonyhurst heliographic coordinates
        lonlat = c.transform_to('heliographic_stonyhurst')

        # construct transformation matrix for the heliographic components of the magnetic field
        # according to Eq (1) in Gary & Hagyard (1990)
        # see Eq (7)(8) in Sun (2013) for implementation

        # b-angle, disk center latitude, in radians
        b = self.keys.CRLT_OBS[0] * (np.pi / 180.)
        # p-angle, negative of CROTA2, in radians
        p = - self.keys.CROTA2[0] * (np.pi / 180.)
        Phi = np.array(lonlat.lon) * (np.pi / 180.)
        Lambda = np.array(lonlat.lat) * (np.pi / 180.)

        sinb = np.sin(b)
        cosb = np.cos(b)
        sinp = np.sin(p)
        cosp = np.cos(p)
        sinphi = np.sin(Phi)
        cosphi = np.cos(Phi)
        sinlam = np.sin(Lambda)
        coslam = np.cos(Lambda)

        k11 = coslam * (sinb * sinp * cosphi + cosp *
                        sinphi) - sinlam * cosb * sinp
        k12 = - coslam * (sinb * cosp * cosphi - sinp *
                          sinphi) + sinlam * cosb * cosp
        k13 = coslam * cosb * cosphi + sinlam * sinb
        k21 = sinlam * (sinb * sinp * cosphi + cosp *
                        sinphi) + coslam * cosb * sinp
        k22 = - sinlam * (sinb * cosp * cosphi - sinp *
                          sinphi) - coslam * cosb * cosp
        k23 = sinlam * cosb * cosphi - coslam * sinb
        k31 = - sinb * sinp * sinphi + cosp * cosphi
        k32 = sinb * cosp * sinphi + sinp * cosphi
        k33 = - cosb * sinphi

        # create output arrays of (Bp,Bt,Br), which is identical to (Bxh, -Byh, Bzh)
        # in Gary & Hagyard (1990), see Appendix in Sun (2013)

        bptr = np.ndarray([self.azimuth[1].header['NAXIS2'],
                           self.azimuth[1].header['NAXIS1'], 3])
        bptr[:, :, 0] = k31 * b_xi + k32 * b_eta + k33 * b_zeta
        bptr[:, :, 1] = k21 * b_xi + k22 * b_eta + k23 * b_zeta
        bptr[:, :, 2] = k11 * b_xi + k12 * b_eta + k13 * b_zeta

        return [lonlat, bptr]



from .prepare_base import PrepareWorker




### demo test ###
def demo_fun_use_Bptr():
    # config
    p_raw_path = r"/media/zander/Data/now/work-inner/sun/LINFF/data/hmi.sharp_cea_720s.377.20110215_020000_TAI.Bp.fits"
    t_raw_path = r"/media/zander/Data/now/work-inner/sun/LINFF/data/hmi.sharp_cea_720s.377.20110215_020000_TAI.Bt.fits"
    r_raw_path = r"/media/zander/Data/now/work-inner/sun/LINFF/data/hmi.sharp_cea_720s.377.20110215_020000_TAI.Br.fits"
    sroot=r"/public1/home/sc81826/temp/out"
    # create object
    pre_worker = PrepareWorker()
    # pre
    pre_worker.prepare_from_fits_Bprt(p_raw_path,t_raw_path,r_raw_path,sroot)
    print("finish demo_fun_use_Bptr_local")

def demo_fun_use_afi():
    save_dir = r"/public1/home/sc81826/temp/out"
    dd=r"/public1/home/sc81826/archive/selAR/2019/hmi.sharp_720s.7334.20190124_124800_TAI"
    a=dd+"/hmi.sharp_720s.7334.20190124_124800_TAI.azimuth.fits"
    f=dd+"/hmi.sharp_720s.7334.20190124_124800_TAI.field.fits"
    i=dd+"/hmi.sharp_720s.7334.20190124_124800_TAI.inclination.fits"
    pre_worker = PrepareWorker()
    pre_worker.prepare_from_fits_afi(a,f,i,save_dir)
    print("finish demo_fun_use_afi_local")

def demo_fun_use_afid():
    save_dir = r"/public1/home/sc81826/temp/out"
    dd=r"/public1/home/sc81826/run/linff-main/run_space/tool/data"
    a=dd+"/hmi.sharp_720s.7300.20180823_173600_TAI.azimuth.fits"
    f=dd+"/hmi.sharp_720s.7300.20180823_173600_TAI.field.fits"
    i=dd+"/hmi.sharp_720s.7300.20180823_173600_TAI.inclination.fits"
    d=dd+"/hmi.sharp_720s.7300.20180823_173600_TAI.disambig.fits"
    pre_worker = PrepareWorker()
    pre_worker.prepare_from_fits_afid(a,f,i,d,save_dir)
    print("finish demo_fun_use_afi_local")


if __name__=="__main__":
    demo_fun_use_Bptr()
    demo_fun_use_afi()
    demo_fun_use_afid()







"""
Purpose:   [1] this project, pynlfff , may not use all function in computer machine
                For example, computer machines may not need the image function
                Therefore, to separate the different function dependencies to achieve mitigation,
                the role of this package is to check whether the package dependencies required
                for different functions are installed


Usage:     This code depends on None
           This code is compatible with python 3.7.x.

Examples:  None Now

Adapted:   ZhaoZhongRui (zhaozhongrui21@mails.ucas.ac.cn) 

"""
import os
import subprocess
import sys

class RCheck():
    """
    check pynlfff some function needed module is install or not
    """

    def __init__(self, version="0.3.x", print_try_log=True,try_install=True):
        """
        version default now, can use for different version test latter,
        print test log can show display or not
        :param version: default now version
        :param print_try_log: default True,show try log
        """
        self.version = version
        self.__data_is_load = False
        self.__requirements_list = []
        self.__load_data()
        self.print_try_log = print_try_log  # print log or not
        self.try_install=try_install

    def __load_data(self):
        """
        load data by version
        :return: None
        """
        if self.version.startswith("0.2"):
            self.__data_v02()
        if self.version.startswith("0.3"):
            self.__data_v03()
        else:
            print("no that version: {}".format(self.version))

    def __data_v02(self):
        """
        load v02 data
        :return: None
        """
        self.__full_list = [
            ["pandas", "", ""],  # 0
            ["astropy", "", ""],  # 1
            ["numpy", "", ""],  # 2
            ["sunpy", "", ""],  # 3
            ["drms", "", ""],  # 4
            ["h5py", "", ""],  # 5
            ["matplotlib", "", ""],  # 6



        ]
        d = self.__full_list
        self.__requirements_base = [d[2]]
        self.__requirements_prepare = [d[0], d[1], d[2], d[3], d[4]]
        self.__requirements_computer = [d[7]]
        self.__requirements_product = [d[5], d[2]]
        self.__requirements_plot = [d[5], d[6], d[2]]

        self.__requirements_list = [
            self.__full_list,  # 0
            self.__requirements_base,  # 1
            self.__requirements_prepare,  # 2
            self.__requirements_computer,  # 3
            self.__requirements_product,  # 4
            self.__requirements_plot,  # 5
        ]
        self.__data_is_load = True

    def __data_v03(self):
        """
        load v02 data
        :return: None
        """
        self.__full_list = [
            ["pandas", "", ""],  # 0
            ["astropy", "", ""],  # 1
            ["numpy", "", ""],  # 2
            ["sunpy", "", ""],  # 3
            ["drms", "", ""],  # 4
            ["h5py", "", ""],  # 5
            ["matplotlib", "", ""],  # 6
            ["psutil", "", ""],  # 7

            ["requests", "", ""],  # 8
            ["bs4", "", ""],  # 9
            ["lxml", "", ""],  # 10
            ["zeep", "", ""],  # 11

            ["wget", "", ""],  # 12
            ["peewee", "", ""],  # 13
            ["vtk", "", ""],  # 14
            
            # ["", "", ""],  # 
        ]
        d = self.__full_list
        self.__requirements_base = [d[2]]
        self.__requirements_prepare = [d[0], d[1], d[2], d[3], d[4]]
        self.__requirements_computer = []
        self.__requirements_product = [d[5], d[2]]
        self.__requirements_plot = [d[5], d[6], d[2], d[14]]
        self.__requirements_download = [d[1],d[3],d[4],d[8], d[9], d[10], d[11]]
        self.__requirements_label = [d[2],d[12],d[13]]


        self.__requirements_list = [
            self.__full_list,  # 0
            self.__requirements_base,  # 1
            self.__requirements_prepare,  # 2
            self.__requirements_computer,  # 3
            self.__requirements_product,  # 4
            self.__requirements_plot,  # 5
            self.__requirements_download, # 6
            self.__requirements_label, # 7
        ]
        self.__data_is_load = True
        
    def h(self):
        """
        print help
        :return: None
        """
        self.help()

    def help(self):
        """
        print help
        :return: None
        """
        help_str = """This module is to test some module, for pynlfff which function you need, is install or not.
    Can use RCheck().check() or RCheck().check(0) for Full module test
    RCheck().check(1) for Base function module test 
    RCheck().check(2) for Prepare function module test 
    RCheck().check(3) for Computer function module test 
    RCheck().check(4) for Product function module test 
    RCheck().check(5) for Plot function module test 
    RCheck().check(6) for Download fits raw data function module test 
    RCheck().check(7) for Label function module test 
    """  # Full Base  Prepare  Computer Product  Plot
        print(help_str)

    def check_one_module_exists_by_try(self, model_name):
        """
        test module is install or not
        :param model_name: str, this is import name
        :return: True or False
        """
        # model_name = 'time'
        result = False
        try:
            log = __import__(model_name)
            if self.print_try_log:
                print(log)
            result = True
        except BaseException as e:
            if self.print_try_log:
                print(e)
        return result

    def check_one_module(self, model_detail_list):
        """
        check one module by module list
        :param model_detail_list: like ["module_name_str","..."],index 0 is name ,other dev latter
        :return: [module_name,exists_str]
        """
        model_name = model_detail_list[0]

        exists_str = "Need Install"
        is_exists = self.check_one_module_exists_by_try(model_name)
        if self.print_try_log:
            print(is_exists)
        if is_exists:
            exists_str = "Exists"

        result = [model_name, exists_str]
        return result

    def check(self, num=0):
        """
        given function index num to check this function's requirements is install or not,
        the num can use RCheck().help() or RCheck().h() to get details.
        :param num: int 0 to 5, 0:Full 1:Base  2:Prepare  3:Computer 4:Product  5:Plot 6:Download raw fits 7:Label
        :return: print result and return test list , error num will return False
        """
        result = False
        if 0 <= num < len(self.__requirements_list):
            check_list = self.__requirements_list[num]
            result_list = []
            if self.print_try_log:
                print("Start Test")
            for check_model_detail_list in check_list:
                this_result = self.check_one_module(check_model_detail_list)
                result_list.append(this_result)
            if self.print_try_log:
                print("Test Finish, Result: {}".format(result_list))
                
            if self.try_install:
                install_fail_list=[]
                for r in result_list:
                    if self.print_try_log:
                        print(r)
                    if r[1]=="Need Install":
                        try:
                            # os.system("{} -m pip install {}".format(sys.executable,r[0]))
                            command="{} -m pip install {}".format(sys.executable,r[0])
                            ret=subprocess.run(command, shell=True)
                            if ret.returncode==0:
                                print("install sucess {}".format(r[0]))
                            else:
                                print("install fail {}".format(r[0]))
                                install_fail_list.append(r)
                        except BaseException as e:
                            print(e)
                if len(install_fail_list)>0:
                    print("=====\ninstall fail: {}".format(install_fail_list))
            result = result_list
        else:
            print("ERROR num: {}\ntry use function RCheck.h() or RCheck.help()".format(num))
        return result


if __name__ == "__main__":
    print("start test")
    # rc=RCheck()
    # RCheck(print_try_log=False).h()
    RCheck(print_try_log=False).check()



# coding=utf-8
"""
Purpose:   [1] run group in cpu

Usage:     This code depends on the psutil
           This code is compatible with python 3.7.x.

Examples:  None Now

Adapted:   ZhaoZhongRui (zhaozhongrui21@mails.ucas.ac.cn) Edit Python (2022.03)

"""

import os
import sys
import multiprocessing
import psutil
import time
import re
from multiprocessing import managers
import socket


def quality_is_ok(file_path):
    """Check whether the file degree angle is greater than 30, 
    greater than 30, or the file is empty, return False, 
    otherwise True, to ensure that the file exists

    Args:
        file_path ([type]): [description]

    Returns:
        [type]: [description]
    """
    if not os.path.exists(file_path):
        result = False
    else:
        with open(file_path, 'r')as f:
            file_data = f.read()
        # print(file_data)
        pattern = r'Angle.*?Degree'  # Pattern String
        string = file_data  # String to match
        match = re.findall(pattern, string)
        # print(match)
        p = r'\d+\.?\d+'
        nums = re.findall(p, str(match))
        if len(nums) == 0:
            result = False
        else:
            result = True
            for num in nums:
                if float(num) > 30:
                    result_n = False
                else:
                    result_n = True
                result = result and result_n
    return result


class QueueManager(managers.BaseManager):

    def __init__(self, model="client", server_config_path="config.ini",server_port=5019):

        self.model = model  # "server" "client"
        if self.model == "server":
            self.init_server_queue()
        self.config_path = server_config_path
        self.this_address = None
        self.port = server_port
        self.get_address()
        authkey = b"1001"
        serializer = 'pickle'
        ctx = None
        print(self.this_address)
        super().__init__(address=self.this_address, authkey=authkey, serializer=serializer, ctx=ctx)
        self.start_run()

    def init_server_queue(self):
        manager = multiprocessing.Manager()
        self.queue_todo1 = manager.Queue()
        self.queue_todo2 = manager.Queue()
        self.queue_todo3 = manager.Queue()
        self.queue_todo4 = manager.Queue()

    def get_address(self):
        if self.model == "server":
            self.get_address_server()
        elif self.model == "client":
            self.get_address_client()

    def get_address_server(self):
        address_ip = "127.0.0.1"
        try:
            address_ip = socket.gethostbyname(socket.gethostname())
        except BaseException as e:
            print(e)
        self.this_address = (str(address_ip), self.port)
        with open(self.config_path, "w+") as f:
            f.write(address_ip)

    def get_address_client(self):
        address_ip = "127.0.0.1"
        if os.path.exists(self.config_path):
            with open(self.config_path, "r+") as f:
                address_ip = f.read()
        self.this_address = (str(address_ip), self.port)

    def start_run(self):
        if self.model == "server":
            self.start_run_server()
        elif self.model == "client":
            self.start_run_client()

    def return_queue_todo1(self):
        return self.queue_todo1

    def return_queue_todo2(self):
        return self.queue_todo2

    def return_queue_todo3(self):
        return self.queue_todo3

    def return_queue_todo4(self):
        return self.queue_todo4

    def start_run_server(self):
        # Register two Queues to the network, callable parameter associated Queue object
        # callale does not handle lambda anonymous functions in win10
        self.register('get_queue_todo1', callable=self.return_queue_todo1)
        self.register('get_queue_todo2', callable=self.return_queue_todo2)
        self.register('get_queue_todo3', callable=self.return_queue_todo3)
        self.register('get_queue_todo4', callable=self.return_queue_todo4)
        # Binding port 5000, how did this 5000 come about? Just the same port in both files!
        # Expose the Queue through the QueueManager
        self.start()

    def start_run_client(self):
        self.register('get_queue_todo1')
        self.register('get_queue_todo2')
        self.register('get_queue_todo3')
        self.register('get_queue_todo4')
        self.connect()


class linff_sim():
    def __init__(self, root_path, shs, shp,mode="server"):
        self.root_path = root_path
        self.root_path_q2sucess=os.path.join(self.root_path,"q2sucess")
        self.root_path_q3sucess=os.path.join(self.root_path,"q3sucess")

        if not os.path.exists(self.root_path_q2sucess):
            os.makedirs(self.root_path_q2sucess)
        if not os.path.exists(self.root_path_q3sucess):
            os.makedirs(self.root_path_q3sucess)
        self.job_1_fail = os.path.join(self.root_path, "fail1.txt")
        self.server_config = os.path.join(self.root_path, "server.txt")
        if not os.path.exists(self.server_config):
            self.run_flag_file = os.path.join(self.root_path,
                                              "server_is_run.flag.{}.{}".format(socket.gethostname(), time.time()))
        else:
            self.run_flag_file = os.path.join(self.root_path,
                                              "worker_is_run.flag.{}.{}".format(socket.gethostname(), time.time()))
        print(self.run_flag_file)
        self.shp = shp
        self.shs = shs
        manager = multiprocessing.Manager()
        self.cpu_list_in_use = manager.list()
        self.is_get_cpu_lock = manager.Lock()
        self.lock_level_2 = manager.Value("i", 0)
        self.level_2_min = 3
        self.lock_level_3 = manager.Value("i", 0)
        self.level_3_min = 3
        if mode=="server":
            self.with_server = True
        else:
            self.with_server = False
        self.init_job_queue()

    def init_job_queue(self):
        try:
            if not os.path.exists(self.server_config): 
                self.with_server = True
                manager = QueueManager("server", self.server_config)   
            else:
                manager = QueueManager("client", self.server_config)
            self.job_todo1_queue = manager.get_queue_todo1()
            self.job_todo2_queue = manager.get_queue_todo2()
            self.job_todo3_queue = manager.get_queue_todo3()
            self.job_todo4_queue = manager.get_queue_todo4()
            self.job_todon_queue_list = [self.job_todo1_queue, self.job_todo2_queue, self.job_todo3_queue,
                                         self.job_todo4_queue]
        except BaseException as e:
            print("init_job_queue error:")
            print(e)

    def server_distribute_job_queue(self, numb):
        self.job_1_todo = os.path.join(self.root_path, "todo1.txt")
        self.job_2_todo = os.path.join(self.root_path, "todo2.txt")
        self.job_3_todo = os.path.join(self.root_path, "todo3.txt")
        self.job_4_todo = os.path.join(self.root_path, "todo4.txt")
        self.job_n_todo = [self.job_1_todo, self.job_2_todo, self.job_3_todo, self.job_4_todo]
        self.__file_if_not_exist_creat_12(self.job_1_todo)
        self.__file_if_not_exist_creat(self.job_2_todo)
        self.__file_if_not_exist_creat(self.job_3_todo)
        self.__file_if_not_exist_creat(self.job_4_todo)
        while self.is_run_now():
            try:
                for i in range(3):
                    if os.path.exists(self.job_n_todo[i]):
                        with open(self.job_n_todo[i], "r+") as f:
                            # f.readlines()
                            this_lines = f.read().splitlines()
                        for job in this_lines:
                            if job.startswith("hmi"):
                                self.job_todon_queue_list[i].put(job)
                        # os.remove(self.job_n_todo[i])
                        os.rename(self.job_n_todo[i], "{}.{}".format(self.job_n_todo[i], time.time()))
                self.server_print_status()
            except BaseException as e:
                print(e)
                self.init_job_queue()
            time.sleep(10 * 60)

    def get_level_lock(self, level):
        if level == 2:
            self.lock_level_2.value = self.lock_level_2.value + 1
        if level == 3:
            self.lock_level_3.value = self.lock_level_3.value + 1

    def release_level_lock(self, level):
        if level == 2:
            self.lock_level_2.value = self.lock_level_2.value - 1
        if level == 3:
            self.lock_level_3.value = self.lock_level_3.value - 1

    def check_level_no_lock(self, level):
        """No lock return true

        Args:
            level (_type_): _description_

        Returns:
            _type_: _description_
        """
        if level == 2:
            if self.lock_level_2.value == -1:  # All tasks completed
                return True
            elif self.lock_level_2.value > self.level_2_min:
                return True
            else:
                return False
        elif level == 3:
            if self.lock_level_3.value == -1:  # All tasks completed
                return True
            elif self.lock_level_3.value > self.level_3_min:
                return True
            else:
                return False

    def server_print_status(self):
        for i in range(4):
            joblist = []
            qz = int(self.job_todon_queue_list[i].qsize())
            for j in range(qz):
                try:
                    job = self.job_todon_queue_list[i].get_nowait()
                    joblist.append(job)
                    self.job_todon_queue_list[i].put(job)
                except BaseException as e:
                    print(e)

            f_path = "{}.status.{}".format(self.job_n_todo[i], time.time())
            with open(f_path, "w+") as f:
                f.write("{}\n{}".format(qz,joblist))

    def client_keep_job_queue(self, numb):
        if numb == 0:
            return False
        self.job_1_todo = os.path.join(self.root_path, "todo1.txt")
        self.job_2_todo = os.path.join(self.root_path, "todo2.txt")
        self.job_3_todo = os.path.join(self.root_path, "todo3.txt")
        self.job_4_todo = os.path.join(self.root_path, "todo4.txt")
        self.job_n_todo = [self.job_1_todo, self.job_2_todo, self.job_3_todo, self.job_4_todo]
        while self.is_run_now():
            try:
                for i in range(3):
                    if os.path.exists(self.job_n_todo[i]):
                        with open(self.job_n_todo[i], "r+") as f:
                            this_lines = f.read().splitlines()
                        for job in this_lines:
                            if job.startswith("hmi"):
                                self.job_todon_queue_list[i].put(job)
                        os.rename(self.job_n_todo[i], "{}.{}".format(self.job_n_todo[i], time.time()))
                self.server_print_status()
            except BaseException as e:
                print(e)
                self.init_job_queue()
            time.sleep(20 * 60 + 3)


    def main_control(self):
        print("start control")
        with open(self.run_flag_file, "a+") as f:
            f.write(str(time.localtime()))

        self.__file_if_not_exist_creat(self.job_1_fail)

        self.run_pool()
        return True

    def run_pool(self):
        print("start pool")
        r1 = 6
        r2 = 5
        r3 = 10
        rf = 0

        if self.with_server:
            q = multiprocessing.Process(target=self.server_distribute_job_queue, args=(1,))
        else:
            q = multiprocessing.Process(target=self.client_keep_job_queue, args=(0,))
        job = []
        for r3_num in range(r3):
            j = multiprocessing.Process(target=self.run_job_n, args=(3, r3_num,))
            job.append(j)
        for r1_num in range(r1):
            j = multiprocessing.Process(target=self.run_job_n, args=(1, r1_num,))
            job.append(j)
        for r2_num in range(r2):
            j = multiprocessing.Process(target=self.run_job_n, args=(2, r2_num,))
            job.append(j)

        for rf_num in range(rf):
            j = multiprocessing.Process(target=self.run_job_free_fix, args=(1, rf_num,))
            job.append(j)

        q.start()
        for jq in job:
            jq.start()
        q.join()
        for jq in job:
            jq.join()

    def append_job_to_file(self, level, value):
        j = level - 1
        job = value
        try:
            self.job_todon_queue_list[j].put(job)
        except BaseException as e:
            print("appendjoberror")
            print(e)
            self.init_job_queue()
        print("job {}".format(job))
        return job

    def get_job(self, level, idn):
        j = level - 1
        job = None
        try:
            if idn > 6:
                if level == 1:
                    t1 = 0
                    while (not self.check_level_no_lock(2))  or t1 > 50:
                        time.sleep(5)
                        t1 = t1 + 1
                    t1 = 0
                    while (not self.check_level_no_lock(3))  or t1 > 50:
                        time.sleep(5)
                        t1 = t1 + 1
            if not self.job_todon_queue_list[j].empty():
                job = self.job_todon_queue_list[j].get()
            else:
                if level == 2:
                    self.lock_level_2.value = -1
                elif level == 3:
                    self.lock_level_3.value = -1
        except BaseException as e:
            print("getjoberror")
            print(e)
            self.init_job_queue()
        print("job {}".format(job))
        return job

    def run_job_n(self, level, id1):
        print("run job {}:{}\n".format(level, id1))

        while self.is_run_now():
            job = self.get_job(level, id1)
            if job is not None:
                # print("==========")
                need_cpu = self.get_need_cpu_num_by_grid(job, level)
                cpu_list = self.get_free_cpu_list(need_cpu)
                run_result = self.run_job_base(job, cpu_list, level)
                if run_result is "ing":
                    print("ing")
                elif run_result is True:
                    self.append_job_to_file(level + 1, job)
                    if level==2:
                        q2flagsucess=os.path.join(self.root_path_q2sucess,"{}.sucess".format(job))
                        with open(q2flagsucess,"w+") as f:
                            f.write("{}".format(time.time()))
                    elif level==3:
                        q3flagsucess=os.path.join(self.root_path_q3sucess,"{}.sucess".format(job))
                        with open(q3flagsucess,"w+") as f:
                            f.write("{}".format(time.time()))
                else:
                    self.append_job_to_file(1, job)
            else:
                time.sleep(5)

    def run_job_free_fix(self, level, idf):
        print("run job f:{}\n".format(idf))
        while self.is_run_now():
            this_level = level  # idf+1
            # done1 = "{}.done".format(self.job_1_todo)
            # if os.path.exists(done1):
            #     this_level = 2
            #     done2 = "{}.done".format(self.job_2_todo)
            #     if os.path.exists(done2):
            #         this_level = 3
            job = self.get_job(level, 0)
            if job is None:
                time.sleep(10)
                continue
            time.sleep(50)
            all_1 = self.get_free_cpu_list_base()
            time.sleep(50)
            all_2 = self.get_free_cpu_list_base()

            cpu_list = list((set(all_1) & set(all_2)) - set(self.cpu_list_in_use))
            if len(cpu_list) > 3:
                for cpu in cpu_list:
                    self.cpu_list_in_use.append(cpu)

                job = self.get_job(this_level, idf)
                run_result = self.run_job_base(job, cpu_list, this_level)
                if run_result is "ing":
                    print("ing")
                elif run_result is True:
                    self.append_job_to_file(this_level + 1, job)
                else:
                    self.append_job_to_file(1, job)

    def get_level(self,job):
        job_name_list = job.split(".")
        num = int(job_name_list[2])
        num_level_1 = num//1000
        num_level_2 = (num//100) % 10
        child_dir= "num_{}{}00_{}{}99".format(num_level_1, num_level_2, num_level_1, num_level_2)
        
        return child_dir

    def run_job_base(self, job, cpu_num_list, level):
        print("start run job: {}  use cpu: {}".format(job, cpu_num_list))
        self.get_level_lock(level)
        work_path = os.path.join(self.root_path,self.get_level(job), job)
        log_path = os.path.join(self.root_path,self.get_level(job), job, "run.log")
        ing_path = os.path.join(self.root_path,self.get_level(job), job, "ing.flag")
        q3flagsucess = os.path.join(self.root_path_q3sucess, "{}.sucess".format(job))
        if os.path.exists(ing_path) or os.path.exists(q3flagsucess) or not os.path.exists(work_path):  # or os.path.exists(done_path):
            to_free_cpu = list(set(self.cpu_list_in_use) - set(cpu_num_list))
            for cpu in to_free_cpu:
                self.cpu_list_in_use.remove(cpu)
            return "ing"
        else:
            with open(ing_path, "w+") as f:
                f.write(str(time.localtime()))

        is_base_file = self.check_base_file_need(job)
        if not is_base_file:
            print("loss file")
            with open(self.job_1_fail, "a+") as f:
                f.write("{}\n".format(job))
            return "ing"
        if level==1:
            self.run_job_qfail_remove(job)
        if len(cpu_num_list) == 1:
            cpus_use = str(cpu_num_list[0])
            run_sh = self.shs
        elif len(cpu_num_list) > 1:
            cpus_use = ""
            for c in cpu_num_list:
                cpus_use = cpus_use + "{},".format(c)
            cpus_use = cpus_use[:-1]
            run_sh = self.shp
        else:
            run_result = False
            return run_result

        try:
            cmd = "/usr/bin/taskset -c {0} {1} {2}  {3} >> {4} ".format(cpus_use, run_sh, work_path, level, log_path)
            returncode = os.system(cmd)  # TAG  subprocess.Popen Will block

            if returncode == 0:
                # run_result = True
                q_f = os.path.join(work_path, "NLFFFquality{}.log".format(level))
                if os.path.exists(q_f):
                    run_result = quality_is_ok(q_f)
                else:
                    self.run_job_qfail_remove(job)
                    run_result = False

            else:
                print("{}{}".format(cmd, returncode))
                # print(returncode)
                run_result = False
        except BaseException as e:
            print("{}{}".format(job, e))
            # print()
            run_result = False

        to_free_cpu = list(set(self.cpu_list_in_use) - set(cpu_num_list))
        for cpu in to_free_cpu:
            self.cpu_list_in_use.remove(cpu)

        if os.path.exists(ing_path):
            os.remove(ing_path)
        self.release_level_lock(level)
        return run_result

    def check_base_file_need(self, job):
        r_path = os.path.join(self.root_path,self.get_level(job), job)
        base_set = {"mask1.dat", "mask2.dat", "mask3.dat", "grid1.ini", "grid2.ini", "grid3.ini",
                    'allboundaries1.dat', 'allboundaries2.dat', 'allboundaries3.dat', 'boundary.ini'}
        this_set = set(os.listdir(r_path))
        if len(base_set - this_set) > 0:
            print(base_set - this_set)
            return False
        else:
            return True

    def run_job_qfail_remove(self, job):
        r_path = os.path.join(self.root_path,self.get_level(job), job)
        # with open(self.job_u_remove_sh, "a+") as f:
        #     f.write("rm -rf {}\n".format(r_path))
        try:
            p_list = ['Energy.log', 'grid2n2.ini', 'NLFFFquality.log',
                      'NLFFFquality3.log', 'step3.log', 'Bout.bin',
                      'Iteration_start.mark', 'prot2.log', 'NLFFFquality2.log',
                      'done1.txt', 'prot1.log', 'NLFFFquality1.log',
                      'prot3.log', 'zboundaries2n2.dat', 'done.txt', 'grid.ini',
                      'done2.txt', 'prot.log',
                      'step.log', 'step2.log', 'allboundaries.dat' , 'B0.bin',
                      'Iteration_stop.mark', 'mask.dat',
                      'step1.log'] #

            for p in p_list:
                p_path = os.path.join(r_path, p)
                if os.path.exists(p_path):
                    os.remove(p_path)
            # shutil.rmtree(r_path)
        except BaseException as e:
            print("{} remove error {}\n".format(job, e))

    def get_need_cpu_num_by_grid(self, job, level):
        use_num = 1
        l_f = os.path.join(self.root_path,self.get_level(job), job, "grid{}.ini".format(level))
        if os.path.exists(l_f):
            with open(l_f, "r") as f:
                lines = f.readlines()
            if len(lines) == 10:
                xyz = int(lines[1]) * int(lines[3]) * int(lines[5])
                use_num = int(xyz / 1024 / 1024 / 5 / 4) + 1
                if use_num != 1:
                    use_num = use_num + 4
        return use_num

    def get_free_cpu_list(self, cpu_need_nums):

        cpu_free_all_list = self.get_free_cpu_list_base()
        while self.is_run_now() and len(cpu_free_all_list) - len(self.cpu_list_in_use) - cpu_need_nums < 0:
            print("cpu {}/{}".format(len(cpu_free_all_list), cpu_need_nums))
            time.sleep(20)
            cpu_free_all_list = self.get_free_cpu_list_base()
        cpu_list = list(set(cpu_free_all_list) - set(self.cpu_list_in_use))
        cpu_list = cpu_list[:cpu_need_nums]
        for cpu in cpu_list:
            self.cpu_list_in_use.append(cpu)
        return cpu_list

    def get_free_cpu_list_base(self):
        result_list = []
        self.is_get_cpu_lock.acquire()
        all_p = psutil.cpu_percent(interval=6, percpu=True)
        for i in range(len(all_p)):
            if all_p[i] < 10:
                result_list.append(i)
        self.is_get_cpu_lock.release()
        return result_list

    def is_run_now(self):
        if os.path.exists(self.run_flag_file):
            result = True
        else:
            result = False
        return result

    def __dir_if_not_exist_creat(self, dir_path):
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

    def __file_if_not_exist_creat(self, file_path):
        father_dir = os.path.dirname(file_path)
        self.__dir_if_not_exist_creat(father_dir)
        if not os.path.exists(file_path):
            with open(file_path, "w+") as f:
                f.write("")

    def __file_if_not_exist_creat_12(self, file_path):
        # os.remove(file_path)
        if not os.path.exists(file_path):
            self.__file_if_not_exist_creat(file_path)
            father_dir = os.path.dirname(file_path)
            file_list = os.listdir(father_dir)
            with open(file_path, "w+") as f:
                for dir_name in file_list:
                    if dir_name.startswith("hmi.sharp_cea_720s."):
                        f.write("{}\n".format(dir_name))


def get_config():
    if len(sys.argv) < 4:
        print("Please enter 3 parameters.")
        result = None
    else:
        data_root = sys.argv[1]  # him folder parent root directory
        shs_path = sys.argv[2]  # sh single-threaded script path
        shp_path = sys.argv[3]  # sh multithreaded script path
        if os.path.exists(data_root) and os.path.exists(shs_path) and os.path.exists(shp_path):
            result = dict(data_root=data_root, shs_path=shs_path, shp_path=shp_path)
        else:
            print("Error in input parameters")
            print(os.path.exists(data_root), os.path.exists(
                shs_path), os.path.exists(shp_path))
            result = None
    return result


def test():
    config_dict = dict()
    config_dict["data_root"] = r"/home/ma-user/work/project/pre_b2/num_0300_0399/"
    config_dict["shs_path"] = r"/home/ma-user/work/project/linff-sim/app_c/compiled.single/multigrid.sh"
    config_dict["shp_path"] = r"/home/ma-user/work/project/linff-sim/app_c/compiled.parallel/multigrid.sh"

    os.environ[
        "OMP_PROC_BIND"] = "true"  # https://support.huaweicloud.com/tngg-kunpenghpcs/kunpenghpcsolution_05_0025.html
    ls = linff_sim(root_path=config_dict["data_root"], shs=config_dict["shs_path"], shp=config_dict["shp_path"])
    ls.main_control()


if __name__ == "__main__":
    config_dict = get_config()
    if config_dict is not None:
        os.environ[
            "OMP_PROC_BIND"] = "true"  # https://support.huaweicloud.com/tngg-kunpenghpcs/kunpenghpcsolution_05_0025.html
        ls = linff_sim(root_path=config_dict["data_root"], shs=config_dict["shs_path"], shp=config_dict["shp_path"])
        ls.main_control()

    # test()

# python3 run.py dataroot sh2 sh3



# coding=utf-8

"""
Purpose:   [1] Read the bin file which is produced by nlfff (Thomas Wiegelmann's)

Usage:     This code depends on the numpy h5py
           The first libraries are in the standard Anaconda distribution.
           h5py can be installed from conda or pip
           This code is compatible with python 3.7.x.

Examples:  None Now

Adapted:   From Thomas Wiegelmann (wiegelmann@mps.mpg.de)'s IDL code to do the same thing (2011.20)
           ZhaoZhongRui (zhaozhongrui21@mails.ucas.ac.cn) & ZhuXiaoShuai(xszhu@bao.ac.cn)  Edit Python Reference Thomas Wiegelmann (2022.03)

"""

import os
import re
import time
import h5py
import numpy as np


class NlfffFile():
    def __init__(self, bin_endian="little", bin_float_size=8):
        """
        given bin_endian and bin_float_size is the computer product bin data format not now read computer format
        :param bin_endian: little or big
        :param bin_float_size: 8 or 4
        """
        if bin_endian == "little":  # for read bin use
            self.endian = "<"
        elif bin_endian == "big":
            self.endian = ">"
        self.float_size = bin_float_size
        if bin_float_size == 8:
            self.float_format = "d"  # double
        elif bin_float_size == 4:
            self.float_format = "f"  # float
        self.data_hdf_dataset_name="Bxyz"

    def check_bin_size_with_grid(self, bin_path, grid_path=None,nx=None,ny=None,nz=None):
        """
        check bin size agree with bin or not
        :param bin_path: the path for Bout.bin or B0.bin,include file's name
        :param grid_path: the path for grid.ini path,include file's name
        :return: True or False
        """
        result = False
        if isinstance(grid_path,str) and os.path.exists(bin_path) and os.path.exists(grid_path):
            bin_xyz_size = self.get_size_from_grid(grid_path)
        elif isinstance(nx,int) and isinstance(ny,int) and isinstance(nz,int):
            bin_xyz_size=[nx,ny,nz]
        else:
            raise Exception("grid_path or nx,ny,nz not given or illegal")
        bin_os_size = os.path.getsize(bin_path)
        norm_size = 3 * self.float_size  # 24 like 24=3*8  three heft and 8byte for double
        bin_com_size = bin_xyz_size[0] * bin_xyz_size[1] * bin_xyz_size[2] * norm_size
        # like 10240 > 105371648 - (140 * 224 * 140 * 24) > 0
        can_delta_max = 10240  # byte
        can_delta_min = 0
        if can_delta_max >= bin_os_size - bin_com_size >= can_delta_min:
            result = True
        return result

    def get_size_from_grid(self, grid_path):
        """
        get size from grid, which is just read grid.ini and re
        :param grid_path:
        :return: list like [nx, ny, nz] or read fail is False
        """
        result = False
        with open(grid_path, "r") as f:
            l = f.read()
            l = re.split('[\n|\t]', l)
            l = list(filter(None, l))
            if len(l) >= 6:
                nx = int(l[1])
                ny = int(l[3])
                nz = int(l[5])
                result = [nx, ny, nz]
        # print(result)
        return result

    def __bin2array_with_nxyz_oneload_alldata_nomap(self, nx, ny, nz, bin_path):
        """
        load all bin data to memory and tran to array
        :param nx: nx size
        :param ny: ny size
        :param nz: nz size
        :param bin_path: the path of bin
        :return: numpy array shape like (3,nx,ny,nz)
        """
        pic = False
        float_format = self.float_format
        endian = self.endian
        np_dtype_str = "{}{}".format(endian, float_format)  # like little '<d'
        with open(bin_path, "rb") as f:
            pic = np.fromfile(
                f,  # which is open
                dtype=np.dtype(np_dtype_str),
                # np_dtype_str eg '<d' little https://numpy.org/doc/stable/reference/arrays.dtypes.html
                count=-1,  # all data
                sep='',  # Binary without spacing
                offset=0
                # The offset from the beginning is 0, note that this is the first operation after opening the file
            ).reshape(
                (3, nx, ny, nz),  # 3*nx*ny*nz
                order='C'  # According to the C language order reshape, that is, the last (nz) changes the fastest
            )
        # time.sleep(100)
        return pic

    def __bin2array_with_nxyz_oneload_alldata_memmap(self, nx, ny, nz, bin_path):
        """
        not load all bin data to memory but use map, and tran to array
        :param nx: nx size
        :param ny: ny size
        :param nz: nz size
        :param bin_path: the path of bin
        :return: numpy array shape like (3,nx,ny,nz)
        """
        pic = False
        float_format = self.float_format
        endian = self.endian
        np_dtype_str = "{}{}".format(endian, float_format)  # eg '<d'
        pic = np.memmap(bin_path,
                        dtype=np.dtype(np_dtype_str),
                        offset=0,
                        shape=(3, nx, ny, nz),
                        order='C')  # https://numpy.org/doc/stable/reference/generated/numpy.memmap.html
        # time.sleep(100)
        return pic

    def read_bin(self, bin_path, grid_path=None,nx=None,ny=None,nz=None, memmap=True):
        """
        read bin to numpy array shape like (3,nx,ny,nz)
        :param bin_path: the path of Bout.bin or B0.bin
        :param grid_path: the path of grid.ini
        :param nx: the int value of nx
        :param ny: the int value of ny
        :param nz: the int value of nz
        :param memmap: True or False,default True, True use np.memmap
        :return: numpy array shape like (3,nx,ny,nz) ,or size is not ok return False
        """
        result = False
        if isinstance(grid_path,str) and os.path.exists(grid_path):
            cz = self.check_bin_size_with_grid(bin_path, grid_path=grid_path)
            nxyz = self.get_size_from_grid(grid_path)
        elif isinstance(nx,int) and isinstance(ny,int) and isinstance(nz,int):
            cz = self.check_bin_size_with_grid(bin_path, nx=nx,ny=ny,nz=nz)
            nxyz=[nx,ny,nz]
        else:
            raise Exception("grid_path or nx,ny,nz not given or illegal")
        if cz:
            if memmap:  # like lazy load
                result = self.__bin2array_with_nxyz_oneload_alldata_memmap(nxyz[0], nxyz[1], nxyz[2], bin_path)
            else:
                result = self.__bin2array_with_nxyz_oneload_alldata_nomap(nxyz[0], nxyz[1], nxyz[2], bin_path)
        return result
    
    def read_bin2(self,head_fileName=None,bout_fileName=None,bout_dir=None):
        if bout_dir is not None:
            head_fileName=os.path.join(bout_dir,'grid3.ini')
            bout_fileName=os.path.join(bout_dir,'Bout.bin')

        with open(head_fileName,'rb') as f:
            f.readline()
            nx=int(f.readline())
            f.readline()
            ny=int(f.readline())
            f.readline()
            nz=int(f.readline())
            f.close()

        nc=3    
        data=np.fromfile(file=bout_fileName,offset=0,count=3*nx*ny*nz,dtype=np.double).reshape(nc,nx,ny,nz)
        print('[bout] shape: %s'%(str(data.shape)))

        return data

    def tran_bin2hdf5(self, bin_path, hdf5_path, grid_path=None,nx=None,ny=None,nz=None, memmap=True, overwrite=True):
        """
        tran nlfff bin data to hdf5 format
        :param bin_path: the path of Bout.bin or B0.bin
        :param grid_path: the path of grid.ini
        :param hdf5_path: the save path of hdf5_path, with .h5 or .hdf5 suffix
        :param memmap: True or False, default True open bin to np array tran use memmap inner
        :param overwrite: True or False, default True, if hdf5 is exists,will overwrite
        :return: tran sucess will return True otherwise return False
        """
        result = False
        data_array = self.read_bin(bin_path=bin_path, grid_path=grid_path,nx=nx,ny=ny,nz=nz, memmap=memmap)
        if data_array is not False:
            result = self.write_hdf5(data_array, hdf5_path, overwrite)
        return result

    def write_hdf5(self, data_array, hdf5_path, overwrite=True):
        """
        write numpy array shape like (3,nx,ny,nz) to hdf5
        :param data_array: numpy array shape like (3,nx,ny,nz)
        :param hdf5_path: the save path of hdf5_path, with .h5 or .hdf5 suffix
        :param overwrite: True or False, default True, if hdf5 is exists,will overwrite
        :return: save sucess will return True otherwise return False
        """
        result = False
        if hdf5_path.endswith(".h5") or hdf5_path.endswith(".hdf5"):
            this_path = os.path.split(hdf5_path)
            if not os.path.exists(this_path[0]):
                os.makedirs(this_path[0])
            if not os.path.exists(hdf5_path) or overwrite:
                # dataset_amount = 1,
                # if dataset_amount == 1:
                with h5py.File(hdf5_path, 'w') as hf:
                    hf.create_dataset(self.data_hdf_dataset_name, data=data_array)
                    result = True
                # elif dataset_amount == 3:
                #     with h5py.File(hdf5_path, 'w') as hf:
                #         hf.create_dataset("Bx", data=data_array[0])
                #         hf.create_dataset("By", data=data_array[1])
                #         hf.create_dataset("Bz", data=data_array[2])
        else:
            raise Exception("hdf5_path need end with .h5 or .hdf5")
        return result

    def read_hdf5(self, hdf5_path):
        """
        read numpy array from hdf5 file
        :param hdf5_path: the path of hdf5 file
        :return: numpy array shape like (3,nx,ny,nz)
                if file not exists , or not end with .h5 .hdf5 , or read fail will return False
        """
        result = False
        if os.path.exists(hdf5_path):
            if hdf5_path.endswith(".h5") or hdf5_path.endswith(".hdf5"):
                with h5py.File(hdf5_path, 'r') as hf:
                    result = hf[self.data_hdf_dataset_name][:]
        else:
            raise Exception("hdf5 file not exists")
        return result


class nlfffFind():
    def __init__(self):
        pass

    def get_hmi_list_from_file(self, file_path):
        """
        get hmi list from ls or tree log
        :param file_path: the path of ls or tree log file
        :return: list like ["hmi.sharp_cea_720s.7334.20190125_013600_TAI","hmi.sharp_cea_720s.7334.20190126_062400_TAI"]
                file not exists or open fail will return False
        """
        result = False
        if os.path.exists(file_path):
            with open(file_path, 'r')as f:
                file_data = f.read()
            # print(file_data)
            pattern = r'hmi.*?_TAI'
            result = re.findall(pattern, file_data)
        return result

    def get_hmi_set_from_file(self, file_path):
        """
        get hmi set from ls or tree log
        :param file_path: the path of ls or tree log file
        :return: set like {"hmi.sharp_cea_720s.7334.20190125_013600_TAI","hmi.sharp_cea_720s.7334.20190126_062400_TAI"}
            file not exists or open fail will return False
        """
        result = False
        job_list = self.get_hmi_list_from_file(file_path)
        if job_list is not False:
            result = set(job_list)
        return result

    def get_level_dir_from_name(self, name):
        """
        given name return is level
        :param name: eg hmi.sharp_cea_720s.7334.20190126_062400_TAI
        :return: eg num_7300_7399
        """
        job = name
        job_name_list = job.split(".")
        num = int(job_name_list[2])
        num_level_1 = num // 1000
        num_level_2 = (num // 100) % 10
        this_dir = "num_{}{}00_{}{}99".format(
            num_level_1, num_level_2, num_level_1, num_level_2)
        return this_dir


if __name__ == "__main__":
    print("test")
    bout_bin_path = r"/home/zander/Desktop/Bout.bin"
    grid_path = r"/home/zander/Desktop/grid3.ini"
    h5_path = r"/home/zander/Desktop/Bxyz.h5"

    r = NlfffFile()

    s = r.get_size_from_grid(grid_path)
    print(s)

    s2 = r.read_bin(bout_bin_path, grid_path)
    s2 = r.read_bin(bout_bin_path,nx=204,ny=392,nz=224)
    print(s2)

    r.tran_bin2hdf5(bout_bin_path, h5_path, grid_path)
    r.tran_bin2hdf5(bout_bin_path, h5_path, nx=204,ny=392,nz=224)

    a=np.random.uniform(-5, 5, size=(3, 100,50,40))
    h5_path= r"/home/zander/Desktop/Bxyz2.h5"
    r.write_hdf5(a,h5_path)
    # time.sleep(100)



# coding=utf-8
"""
Purpose:   [1] prepare Bp Bt Br or azimuth field inclination [disambig] for nlfff computer which is version 5 by Thomas Wiegelmann (wiegelmann@mps.mpg.de)

Usage:     This code depends on the pandas  numpy astropy and disambiguation
           The first two libraries are in the standard Anaconda distribution.
           astropy can be installed from conda or pip
           The disambiguation library can be obtained from https: https://github.com/mbobra/SHARPs/blob/master/disambiguation.py
           This code is compatible with python 3.7.x.

Examples:  None Now

Adapted:   From Thomas Wiegelmann (wiegelmann@mps.mpg.de)'s IDL code to do the same thing (2011.20)
           ZhaoZhongRui (zhaozhongrui21@mails.ucas.ac.cn) & ZhuXiaoShuai(xszhu@bao.ac.cn) Edit Python code From Thomas Wiegelmann (2022.03)
"""

import os
import pandas
from astropy.io import fits
import numpy
from . import disambiguation

class PrepareWorker():
    
    def __init__(self,mu=0.1,nd=0,nue=0.00100000,boundary=0):
        self.mu = mu
        self.nd = nd
        self.nue = nue
        self.boundary=boundary


    def read_fits_hdu1_data(self,f_path):
        """read fits and return hudl[1].data

        Args:
            f_path (str): path of fits

        Returns:
            _type_: hudl[1].data
        """
        with fits.open(f_path) as hdul:
            data = hdul[1].data
        return data

    def write_boundary_to_ini(self,save_path):
        """write boundary to ini, which write Some constants to given path

        Args:
            save_path (str): path of save boundary with file name, eg /xxx/grid3.ini

        Returns:
            _type_: Success returns True, failure returns False
        """
        result=False
        grid_str = "nue\n\t{0}\nboundary\n\t{1}\nMask {2}\n".format(self.nue,self.boundary,"B_T/max(B_T)")
        file_path = save_path
        with open(file_path, mode='w', encoding='utf-8') as file_obj:
            file_obj.write(grid_str)
            result = True
        return result
    
    def write_nxy_to_grid(self,nx,ny,nz,save_path):
        """
        write grid.ini
        :param nx: nx
        :param ny: ny
        :param ny: nz
        :param save_path: save_path[1] should store the full save path of grid1.ini,grid2.ini with filename and suffix
        :return: Success returns True, failure returns False
        """
        result=False
        grid_str = "nx\n\t{0}\nny\n\t{1}\nnz\n\t{2}\nmu\n\t{3}\nnd\n\t{4}".format(nx,ny,nz,self.mu,self.nd)
        file_path = save_path
        with open(file_path, mode='w', encoding='utf-8') as file_obj:
            file_obj.write(grid_str)
            result = True
        return result
    
    def write_errormask_to_maskdat(self,save_path,deal_type=2,nx=0,ny=0,b3dx=None,b3dy=None):
        """[summary]

        Args:
            save_path ([path]): save path of mask
            deal_type (int, optional): error mask Algorithms used, Defaults to 2 is B_T/max(B_T), set 1 is keep all 1 in error mask
            nx (int, optional): nx value. Defaults to 0.
            ny (int, optional): ny value. Defaults to 0.
            b3dx ([2d array], optional): b3dx. Defaults to None.
            b3dy ([2d array], optional): b3dy. Defaults to None.

        Returns:
            [bool]: Success returns True, failure returns False
        """
        run_result = False
        mask_result_data=None
        if deal_type==1:
            nxny=nx*ny
            mask_result_data = numpy.ones((nxny,1))
        elif deal_type==2:
            if b3dx is not None and b3dy is not None:
                # print(bx)
                # print(type(bx)) # <class 'numpy.ndarray'>
                # btrans=sqrt(bx^2+by^2)
                # maxBT=max(btrans)
                # mask_result_data=btrans/maxBT
                x_array = b3dx.flatten('F')  # flatten in column-major (Fortran- style) , Because the original idl code is, first fix y, then change x
                y_array = b3dy.flatten('F')
                btrans=numpy.sqrt(x_array**2+y_array**2)
                maxBT=numpy.max(btrans)
                mask_result_data=btrans/maxBT
        if mask_result_data is not None:
            file_path = save_path
            numpy.savetxt(file_path,mask_result_data,fmt='%.05f')
            run_result = True
            
        return run_result
            


    def write_b3dxyz_to_allboundaries(self,dx,dy,dz,save_path):
        """
        write allboundaries
        :param dx: dx
        :param dy: dy
        :param dz: dz
        :param save_path: save_path should store the full save path with allboundaries.dat with filename and suffix
        :return: None
        """
        x_array = dx.flatten('F')  # flatten in column-major (Fortran- style) , Because the original idl code is, first fix y, then change x
        y_array = dy.flatten('F')  
        z_array = dz.flatten('F')
        xyz = numpy.vstack([x_array, y_array, z_array])
        xyz = xyz.flatten('F')
        # file_path = os.path.join(save_path_list,'allboundaries.dat')
        file_path = save_path
        numpy.savetxt(file_path, xyz, fmt="%.6f", delimiter="\n")  # https://numpy.org/doc/stable/reference/generated/numpy.savetxt.html


    def read_b3dxyz_from_array_Bptr(self,Bp,Bt,Br):
        n_b3dx = Bp.shape
        n_b3dy = Bt.shape
        n_b3dz = Br.shape
        b3dx=Bp
        b3dy=-Bt
        b3dz=Br
        if (n_b3dx == n_b3dy) and (n_b3dx == n_b3dz):  # Check legality
            nx = n_b3dx[0]
            ny = n_b3dx[1]
        else:
            nx = 0
            ny = 0
        return b3dx,b3dy,b3dz,nx,ny

    def change_b3dxyz_multiple_4(self, b3dx, b3dy, b3dz, nx, ny):
        nx_m=nx%4
        ny_m=ny%4
        if nx_m != 0:
            nx_s = 1
            nx_e = nx - nx_m + 1
        else:
            nx_s = 0
            nx_e = nx
        if ny_m != 0:
            ny_s = 1
            ny_e = ny - ny_m + 1
        else:
            ny_s = 0
            ny_e = ny
        b3dx = b3dx[nx_s:nx_e,ny_s:ny_e]
        b3dy = b3dy[nx_s:nx_e,ny_s:ny_e]
        b3dz = b3dz[nx_s:nx_e,ny_s:ny_e]
        nx = nx - nx_m
        ny = ny - ny_m
        # print(type(b3dx)) # <class 'numpy.ndarray'>
        # print(b3dx.shape)
        # print(nx,ny)
        nz_not_multiple_4=int((nx+ny)*3/8)
        nz_add_to_4=int(4-nz_not_multiple_4%4)
        nz=nz_not_multiple_4+nz_add_to_4
        return b3dx, b3dy, b3dz, nx, ny, nz

    def prepare_from_fits_Bprt(self,Bp_path, Bt_path, Br_path,save_dir):
        """Generate preprocessed data from Bp.bits,Bt.bits,Br.bits

        Args:
            Bp_path (str): path of Bp.fits
            Bt_path (str): path of Bt.fits
            Br_path (str): path of Br.fits
            save_dir (str): path to save pre file

        Returns:
            _type_: True
        """
        bp_array = self.read_fits_hdu1_data(Bp_path)
        bt_array = self.read_fits_hdu1_data(Bt_path)
        br_array = self.read_fits_hdu1_data(Br_path)
        b3dx, b3dy, b3dz, nx, ny = self.read_b3dxyz_from_array_Bptr(bp_array, bt_array, br_array)
        self.prepare_from_array_to_three(b3dx, b3dy, b3dz, nx, ny, save_dir)
        return True
        
    def base_rebin_half(self,m):
        """average pooling

        Args:
            m (2d array): 2d array to average pooling, Get the values of the four points and take the average

        Returns:
            _type_: 2d array which has polling
        """
        m11=m[0::2,0::2]  # Get the values of the four points
        m12=m[0::2,1::2]  # The front indicates rows and the back indicates columns. 0::2 means intercepted in steps of 2 from 0, which is an odd number. 
        m21=m[1::2,0::2]
        m22=m[1::2,1::2]
        m_new=(m11+m12+m21+m22)/4 # can do so provided that m is an even number of rows and columns, otherwise the intercepted shapes are not the same and cannot be added together i.e. the elements in each of the four field positions are averaged into a new element and the fields do not overlap
        return m_new
        
    def change_b3dxyz_rebin(self,b3dx, b3dy, b3dz, nx, ny, nz):
        b3dx = self.base_rebin_half(b3dx)
        b3dy = self.base_rebin_half(b3dy)
        b3dz = self.base_rebin_half(b3dz)
        nx = int(nx/2)
        ny = int(ny/2)
        nz = int(nz/2) 
        return b3dx, b3dy, b3dz, nx, ny, nz
    
    
    def prepare_from_array_to_three(self,b3dx, b3dy, b3dz, nx, ny, save_dir):
        if not os.path.isdir(save_dir):
            print("mkdir {}".format(save_dir))
            os.makedirs(save_dir)
        # grid3
        b3dx, b3dy, b3dz, nx, ny, nz= self.change_b3dxyz_multiple_4(b3dx, b3dy, b3dz, nx, ny)
        save_path = os.path.join(save_dir,"allboundaries3.dat")
        self.write_b3dxyz_to_allboundaries(b3dx, b3dy, b3dz,save_path)
        save_path = os.path.join(save_dir,"grid3.ini")
        self.write_nxy_to_grid(nx, ny, nz, save_path)
        save_path = os.path.join(save_dir,"mask3.dat")
        self.write_errormask_to_maskdat(save_path,nx=nx,ny=ny,b3dx=b3dx,b3dy=b3dy)
        # grid2
        b3dx, b3dy, b3dz, nx, ny, nz= self.change_b3dxyz_rebin(b3dx, b3dy, b3dz, nx, ny, nz)
        save_path = os.path.join(save_dir,"allboundaries2.dat")
        self.write_b3dxyz_to_allboundaries(b3dx, b3dy, b3dz,save_path)
        save_path = os.path.join(save_dir,"grid2.ini")
        self.write_nxy_to_grid(nx, ny, nz, save_path)
        save_path = os.path.join(save_dir,"mask2.dat")
        self.write_errormask_to_maskdat(save_path,nx=nx,ny=ny,b3dx=b3dx,b3dy=b3dy)
        # grid1
        b3dx, b3dy, b3dz, nx, ny, nz= self.change_b3dxyz_rebin(b3dx, b3dy, b3dz, nx, ny, nz)
        save_path = os.path.join(save_dir,"allboundaries1.dat")
        self.write_b3dxyz_to_allboundaries(b3dx, b3dy, b3dz,save_path)
        save_path = os.path.join(save_dir,"grid1.ini")
        self.write_nxy_to_grid(nx, ny, nz, save_path)
        save_path = os.path.join(save_dir,"mask1.dat")
        self.write_errormask_to_maskdat(save_path,nx=nx,ny=ny,b3dx=b3dx,b3dy=b3dy)
        save_path = os.path.join(save_dir,"boundary.ini")
        self.write_boundary_to_ini(save_path)
        # allboundaries1.dat  grid1.ini  mask1.dat  boundary.ini


    def prepare_from_fits_afi(self,a_raw_path,f_raw_path,i_raw_path,save_dir):
        """
        Generate preprocessed data from azimuth.bits,field.bits,inclination.bits
        :param a_raw_path: azimuth.fits File path (need to include file name)
        :param f_raw_path: field.fits File path (need to include file name)
        :param i_raw_path: inclination.fits File path (need to include file name)
        :param save_dir: pre save dir
        :return: None
        """
        a_hdul = fits.open(a_raw_path)  # Note do not use with open assignment return, otherwise only the header data is retained, there are no two data fields, the default only cache the first header
        a_hdul.verify('silentfix')
        # 'silentfix' fixes and no warnings 'fix' fixes but prints warnings   https://github.com/astropy/astropy/blob/main/docs/io/fits/usage/verification.rst
        dict_header = dict(a_hdul[1].header)
        data_keys_frame_header = pandas.DataFrame([dict_header])
        f_hdul = fits.open(f_raw_path)
        i_hdul = fits.open(i_raw_path)
        data_object = disambiguation.CoordinateTransform(a_hdul, f_hdul, i_hdul, data_keys_frame_header)
        latlon, bptr = disambiguation.CoordinateTransform.ccd(data_object)
        a_hdul.close()
        f_hdul.close()
        i_hdul.close()
        b3dx, b3dy, b3dz, nx, ny = self.read_b3dxyz_from_array_Bptr(bptr[:, :, 0],bptr[:, :, 1],bptr[:, :, 2])
        self.prepare_from_array_to_three(b3dx, b3dy, b3dz, nx, ny, save_dir)


    def prepare_from_fits_afid(self,a_raw_path,f_raw_path,i_raw_path,d_raw_path,save_dir):
        """
        Generate preprocessed data from  azimuth.fits,field.fits,inclination.fits
        :param a_raw_path: azimuth.fits  File path (need to include file name)
        :param f_raw_path: field.fits  File path (need to include file name)
        :param i_raw_path: inclination.fits  File path (need to include file name)
        :param d_raw_path: disambig.fits  File path (need to include file name)
        :param save_dir: pre save dir
        :return: None
        """
        a_hdul = fits.open(a_raw_path)
        a_hdul.verify('silentfix')
        dict_header = dict(a_hdul[1].header)
        data_keys_frame_header = pandas.DataFrame([dict_header])
        f_hdul = fits.open(f_raw_path)
        i_hdul = fits.open(i_raw_path)
        d_hdul = fits.open(d_raw_path)
        basic_obj = disambiguation.Basic(recordset=None,method=2)  
        a_hdul_have_d = disambiguation.Basic.perform_disambiguation(basic_obj,a_hdul,d_hdul)
        data_object = disambiguation.CoordinateTransform(a_hdul_have_d, f_hdul, i_hdul, data_keys_frame_header)
        latlon, bptr = disambiguation.CoordinateTransform.ccd(data_object)
        a_hdul.close()
        f_hdul.close()
        i_hdul.close()
        d_hdul.close()
        b3dx, b3dy, b3dz, nx, ny = self.read_b3dxyz_from_array_Bptr(bptr[:, :, 0],bptr[:, :, 1],bptr[:, :, 2])
        self.prepare_from_array_to_three(b3dx, b3dy, b3dz, nx, ny, save_dir)


    def prepare_from_fits_afid_online(self,save_dir):
        """
        Download data online and process it, just for demonstration, not for actual use
        :param save_dir: pre save dir
        :return:
        """
        ### download azimuth field inclination disambiguation

        ## to Bp Bt Br
        # fetch the data from JSOC by providing a recordset specification and a disambiguation method
        query_info = disambiguation.Basic('hmi.sharp_720s[377][2011.02.15_00:00:00]', 2)
        keys, azimuth, field, inclination, disambig = disambiguation.Basic.get_data(query_info)

        # disambiguate the azimuthal component of the magnetic field
        disambiguated_azimuth = disambiguation.Basic.perform_disambiguation(query_info, azimuth, disambig)

        # construct the field vector in spherical coordinate components on the CCD grid
        data_object = disambiguation.CoordinateTransform(disambiguated_azimuth, field, inclination, keys)
        # print(data_object)
        latlon, bptr = disambiguation.CoordinateTransform.ccd(data_object)

        b3dx, b3dy, b3dz, nx, ny = self.read_b3dxyz_from_array_Bptr(bptr[:, :, 0],bptr[:, :, 1],bptr[:, :, 2])
        self.prepare_from_array_to_three(b3dx, b3dy, b3dz, nx, ny, save_dir)
        print("pre demo_fun_use_afid_online")



### demo test ###
def demo_fun_use_Bptr():
    # config
    p_raw_path = r"/media/zander/Data/now/work-inner/sun/LINFF/data/hmi.sharp_cea_720s.377.20110215_020000_TAI.Bp.fits"
    t_raw_path = r"/media/zander/Data/now/work-inner/sun/LINFF/data/hmi.sharp_cea_720s.377.20110215_020000_TAI.Bt.fits"
    r_raw_path = r"/media/zander/Data/now/work-inner/sun/LINFF/data/hmi.sharp_cea_720s.377.20110215_020000_TAI.Br.fits"
    sroot=r"/public1/home/sc81826/temp/out"
    # create object
    pre_worker = PrepareWorker()
    # pre
    pre_worker.prepare_from_fits_Bprt(p_raw_path,t_raw_path,r_raw_path,sroot)
    print("finish demo_fun_use_Bptr_local")

def demo_fun_use_afi():
    save_dir = r"/public1/home/sc81826/temp/out"
    dd=r"/public1/home/sc81826/archive/selAR/2019/hmi.sharp_720s.7334.20190124_124800_TAI"
    a=dd+"/hmi.sharp_720s.7334.20190124_124800_TAI.azimuth.fits"
    f=dd+"/hmi.sharp_720s.7334.20190124_124800_TAI.field.fits"
    i=dd+"/hmi.sharp_720s.7334.20190124_124800_TAI.inclination.fits"
    pre_worker = PrepareWorker()
    pre_worker.prepare_from_fits_afi(a,f,i,save_dir)
    print("finish demo_fun_use_afi_local")

def demo_fun_use_afid():
    save_dir = r"/public1/home/sc81826/temp/out"
    dd=r"/public1/home/sc81826/run/linff-main/run_space/tool/data"
    a=dd+"/hmi.sharp_720s.7300.20180823_173600_TAI.azimuth.fits"
    f=dd+"/hmi.sharp_720s.7300.20180823_173600_TAI.field.fits"
    i=dd+"/hmi.sharp_720s.7300.20180823_173600_TAI.inclination.fits"
    d=dd+"/hmi.sharp_720s.7300.20180823_173600_TAI.disambig.fits"
    pre_worker = PrepareWorker()
    pre_worker.prepare_from_fits_afid(a,f,i,d,save_dir)
    print("finish demo_fun_use_afi_local")


if __name__=="__main__":
    demo_fun_use_Bptr()
    demo_fun_use_afi()
    demo_fun_use_afid()







"""
Purpose:   [1] plot the nlfff product (hdf or bin format) to picture

Usage:     This code depends on the numpy vtk
           The first librarie is in the standard Anaconda distribution.
           The h5py matplotlib library can be obtained from pip
           This code is compatible with python 3.7.x.

Examples:  None Now

Adapted:   Xinze Zhang (zhangxinze17@mails.ucas.ac.cn) Edit Python (2022)

"""

import os
import numpy as np
from PIL import Image,ImageDraw
import vtk
from vtkmodules.util.numpy_support import numpy_to_vtk,numpy_to_vtkIdTypeArray


def hmiPreprocess(data):
    return (np.clip(data,-200,200)+200)/400*255

def genVtkImage(data):
    nz,ny,nx=data.shape
    volume=numpy_to_vtk(data.flatten())
    vtkImage=vtk.vtkImageData()
    vtkImage.SetDimensions(nx,ny,nz)
    vtkImage.GetPointData().SetScalars(volume)
    return vtkImage

def genVolume(id=3):
    volumeMapper=vtk.vtkFixedPointVolumeRayCastMapper()

    volumeProperty=vtk.vtkVolumeProperty()
    volumeProperty.ShadeOn()

    opacityTransferFunction = vtk.vtkPiecewiseFunction()
    opacityTransferFunction.AddPoint(0, 1.0)
    opacityTransferFunction.AddPoint(127.5, 0.0)
    opacityTransferFunction.AddPoint(255,1.0)
    volumeProperty.SetScalarOpacity(opacityTransferFunction)

    volumeColor=vtk.vtkColorTransferFunction()
    for idx in range(0,256):
        if id==0:
            volumeColor.AddRGBPoint(idx,idx/255,0,0)
        elif id==1:
            volumeColor.AddRGBPoint(idx,0,idx/255,0)
        elif id==2:
            volumeColor.AddRGBPoint(idx,0,0,idx/255)
        else:
            volumeColor.AddRGBPoint(idx,*[idx/255]*3)

    volumeProperty.SetColor(volumeColor)

    volume=vtk.vtkVolume()
    volume.SetMapper(volumeMapper)
    volume.SetProperty(volumeProperty)

    return volume,volumeMapper

def genVolumeRen(axis_idx=3):
    ren= vtk.vtkRenderer()
    ren.SetBackground(0.0, 0.0, 0.0)  
    ren.SetGradientBackground(1)
    
    ren.volume,ren.volumeMapper=genVolume(axis_idx)
    ren.AddVolume(ren.volume)

    return ren

def genCutPlaneRen():
    ren = vtk.vtkRenderer()
    ren.SetBackground(0.0, 0.0, 0.0)  
    ren.SetGradientBackground(1)

    cutPlane=vtk.vtkPlane()
    cutPlane.SetOrigin( 0.0, 0.0, 0.001*2000 )
    cutPlane.SetNormal( 0.0, 0.0, 1.0 )
    cutter=vtk.vtkCutter()
    cutter.SetCutFunction( cutPlane )

    ren.cutter=cutter

    cutterMapper=vtk.vtkPolyDataMapper()
    cutterMapper.SetInputConnection( cutter.GetOutputPort() )
    cutterMapper.SetScalarRange(0, 255.0)
  
    colorTable = vtk.vtkLookupTable()
    colorTable.SetNumberOfTableValues(256)
    for idx in range(0,256):
        colorTable.SetTableValue(idx,*[idx/255]*3,1.0)
    cutterMapper.SetLookupTable(colorTable)

    cutterActor=vtk.vtkActor()
    cutterActor.SetMapper( cutterMapper )

    ren.AddActor(cutterActor)

    return ren,cutPlane

def genAuxComp():
    camera=vtk.vtkCamera()

    axesActor=vtk.vtkAxesActor()
    axesActor.SetPosition(0, 0, 0)
    axesActor.SetTotalLength(2, 2, 2)
    axesActor.SetShaftType(0)
    axesActor.SetCylinderRadius(0.02)

    cubeAxesActor = vtk.vtkCubeAxesActor()
    cubeAxesActor.SetCamera(camera)
    cubeAxesActor.SetFlyMode(3)
    cubeAxesActor.SetXTitle("x")
    cubeAxesActor.SetYTitle("y")
    cubeAxesActor.SetZTitle("z")
    cubeAxesActor.SetUse2DMode(1)
    cubeAxesActor.SetZAxisVisibility(1)
    cubeAxesActor.SetDrawXInnerGridlines(1)
    cubeAxesActor.SetDrawYInnerGridlines(1)
    cubeAxesActor.SetDrawZInnerGridlines(1)

    return camera,axesActor,cubeAxesActor

def setRen(ren,viewport,camera,cubeAxesActor):
    ren.SetViewport(viewport)
    ren.AddActor(cubeAxesActor)
    ren.SetActiveCamera(camera)
    ren.ResetCamera()

def genImg(data,percentage,value_name='bz',axis_name='z',withFn=False):
    nx,ny,nz=data.shape
    if axis_name=='x':
        x=int(nx*percentage/100.0)
        x=0 if x<0 else (x if x<nx else nx-1)
        fn,img='%s_x_%d_%d'%(value_name,percentage,x),Image.fromarray(data[x,:,:]).transpose(Image.ROTATE_90)
    elif axis_name=='y':
        y=int(ny*percentage/100.0)
        y=0 if y<0 else (y if y<ny else ny-1)
        fn,img='%s_y_%d_%d'%(value_name,percentage,y),Image.fromarray(data[:,y,:]).transpose(Image.ROTATE_90)
    else:
        z=int(nz*percentage/100.0)
        z=0 if z<0 else (z if z<nz else nz-1)
        fn,img='%s_z_%d_%d'%(value_name,percentage,z),Image.fromarray(data[:,:,z]).transpose(Image.FLIP_TOP_BOTTOM)

    if withFn:
        draw = ImageDraw.Draw(img)
        draw.text((4,4), fn, fill='black')

    return fn,img

def main_draw_3d_body(dataBout,out_dir,axis_name=['bx','by','bz','bnorm2'],bar="auto"):
    # dataBout=pic4d.copy()
    # out_dir="/home/zander/Desktop/draw"
    dataBout=hmiPreprocess(dataBout)
    dataBout=np.concatenate((dataBout,np.expand_dims(np.linalg.norm(dataBout,axis=0,ord=2)/3,axis=0)),axis=0)
    axis_num,nx,ny,nz=dataBout.shape
    
    screenSizeX=1600
    screenSizeY=1600  

    for axis_idx in range(0,axis_num):
        if bar=="auto":
            ren0=genVolumeRen(axis_idx)
        else:
            ren0=genVolumeRen(3)
        camera,axesActor,cubeAxesActor=genAuxComp()
        setRen(ren0,[0.0, 0.0, 1.0, 1.0],camera,cubeAxesActor)

        textActor = vtk.vtkTextActor()
        textActor.GetTextProperty().SetColor(1.0, 0.0, 0.0)
        textActor.GetTextProperty().SetFontSize(32)

        renWin = vtk.vtkRenderWindow()
        renWin.OffScreenRenderingOn()
        renWin.AddRenderer(ren0)
        renWin.SetSize(screenSizeX, screenSizeY)
        renWin.Render()

        vtkWriter = vtk.vtkJPEGWriter()
        windowto_image_filter = vtk.vtkWindowToImageFilter()
        vtkWriter.SetInputConnection(windowto_image_filter.GetOutputPort())
        windowto_image_filter.SetInputBufferTypeToRGBA()
        windowto_image_filter.SetInput(renWin)

        cubeAxesActor.SetBounds(0,nx*1.2,0,ny*1.2,0,nz*1.2)
        cubeAxesActor.SetXAxisRange(0, nx*1.2)
        cubeAxesActor.SetYAxisRange(0, ny*1.2)
        cubeAxesActor.SetZAxisRange(0, nz*1.2)

        camera.SetViewUp(0, 0, 1)
        camera.SetFocalPoint (0, 0, 0)
        camOffset=0.0
        camPos=max([nx,ny,nz])*4.0
        camera.SetPosition (camPos+camOffset, camPos-camOffset, camPos)
        camera.SetFocalPoint (camOffset, -camOffset, camPos*0.1)
        camera.ComputeViewPlaneNormal()


        data=dataBout[axis_idx].swapaxes(0, 2)
        vtkImagenlfff3D=genVtkImage(data)
        ren0.volumeMapper.RemoveAllInputs()
        ren0.volumeMapper.SetInputData(vtkImagenlfff3D)
        ren0.ResetCamera()
        renWin.Render()
        windowto_image_filter.Modified()
        outFn=os.path.join(out_dir,axis_name[axis_idx]+'_cube.jpg')
        print(outFn)
        vtkWriter.SetFileName(outFn)
        vtkWriter.Write()




import os

def run(sh_path,data_path,grid_level):
    if os.path.exists(sh_path) and os.path.exists(data_path) and grid_level in [1,2,3]:
        cmd="{} {} {}".format(sh_path,data_path,grid_level)
        os.system(cmd)
    else:
        print("check os.path.exists(sh_path) and os.path.exists(data_path) and grid_level in [1,2,3]")
        print(os.path.exists(sh_path) , os.path.exists(data_path) , grid_level in [1,2,3])



# coding=utf-8
"""
Purpose:   [1] check nlfff product quality


Usage:     This code depends on the os re
           The first two libraries are in the standard distribution.
           This code is compatible with python 3.7.x.

Examples:  None Now

Adapted:   ZhaoZhongRui (zhaozhongrui21@mails.ucas.ac.cn) Edit Python (2022.03)

"""

import os
import re


class QualityCheck():
    """
    check nlfff quality
    """

    def __init__(self, quality_max=30):
        self.quality_ok_value_max = quality_max  # 30

    def check_quality_from_bin(self):
        """
        check from bin
        :return:
        """
        # TODO
        return False

    def check_quality_from_log(self, file_path):
        """
        Detect whether the file degree angle is greater than 30,
        greater than 30 or the file is empty, return False,
        otherwise True
        :param file_path: NLFFFquality3.log path
        :return:
        """
        result = False
        if os.path.exists(file_path):
            with open(file_path, 'r')as f:
                file_data = f.read()
            # print(file_data)
            pattern = r'Angle.*?Degree'
            match = re.findall(pattern, file_data)
            # print(match)
            p = r'\d+\.?\d+'
            nums = re.findall(p, str(match))
            if len(nums) == 0:
                result = False
            else:
                result = True
                for num in nums:
                    if float(num) > self.quality_ok_value_max:
                        result_n = False
                    else:
                        result_n = True
                    result = result and result_n  # ensure every value is ok
        return result



# coding=utf-8
"""
Purpose:   [1] download data Bp Bt Br from jsoc
           [2] Demonstrate sunpy for jsoc basic usage,
               Officially already has a good package, this is just a demonstration of basic use

Usage:     This code depends on the requests bs4 lxml zeep drms sunpy astropy
           They can be installed from conda or pip
           This code is compatible with python 3.7.x.

Examples:  None Now

Adapted:   ZhaoZhongRui (zhaozhongrui21@mails.ucas.ac.cn) Edit Python code From Thomas Wiegelmann (2022.03)
"""
# This code is compatible with python 3.7.x.
# python3 -m pip install requests
# python3 -m pip install bs4  # https://stackoverflow.com/questions/11783875/importerror-no-module-named-bs4-beautifulsoup
# python3 -m pip install lxml
# python3 -m pip install zeep
# python3 -m pip install drms  # https://docs.sunpy.org/projects/drms/en/stable/
# python3 -m pip install sunpy
# python3 -m pip install astropy
from sunpy.net import Fido, attrs as a
import astropy.units as u
from concurrent.futures import ProcessPoolExecutor
import os


class DownloadJsoc():

    def __init__(self):

        self.is_print_log = True  # or False not print log

        self.mail_address_list = []  # eg ["demo@demo.com","demo2@demo.com"]
        #  You need to go to the following location URL to register the mailbox filled in the following list,
        #  so that when multi-threaded, it will not affect the download, because a user can only request once at a time
        #  http://jsoc.stanford.edu/ajax/exportdata.html
        # ---
        #  Note that this email address has a service period (it may be two months),
        # beyond which you need to re-register on the official website
        #  Note that the number of requests or downloads of the same ip is also limited,
        # otherwise the ip may be blocked, you need to apply for unblocking,
        # you may also automatically unblock after a long time, it is recommended that a small number of requests,
        # a single search request a large amount of data

        self.data_save_root_path = None  # eg r"C:\Users\Zander\PycharmProjects\pynlfff\pyproduct\test"

    def download_one_by_time_point(self, timepoint, harpnum, email=None, save_path=None):
        """
        Download the file based on the point in time
        :param timepoint: eg "2010-01-01T00:00:00.000"
        :param harpnum: eg "220"
        :param email: eg "eg@eg.com" register at http://jsoc.stanford.edu/ajax/exportdata.html
        :param save_path: str path
        :return: None
        """
        result = False
        try:
            if email is None:
                if isinstance(self.mail_address_list, list) and len(self.mail_address_list) > 0:
                    email = self.mail_address_list[0]
            if save_path is None:
                if os.path.exists(self.data_save_root_path):
                    save_path = self.data_save_root_path

            search_tuple = (
                a.Time(timepoint, timepoint),  # set same start time and end time is like to download this time point

                a.jsoc.Series("hmi.sharp_cea_720s"),
                # choose series  http://jsoc.stanford.edu/JsocSeries_DataProducts_map.html

                a.jsoc.Notify(email),  # a.jsoc.Notify("eg@gmail.com"),

                a.jsoc.Segment("Bp"),
                a.jsoc.Segment("Bt"),
                a.jsoc.Segment("Br"),

                a.jsoc.PrimeKey("HARPNUM", str(harpnum)),

                # a.jsoc.Keyword("LON_MIN") > -70, # JSOC keyword filtering with Fido  https://docs.sunpy.org/en/stable/whatsnew/3.1.html?highlight=t_rec#jsoc-keyword-filtering-with-fido
                # a.jsoc.Keyword("LON_MAX") < 70, # Note You can only filter numbers, not strings

                # a.Sample(96 * u.min), # time delta, need set time range before
            )

            results = Fido.search(*search_tuple)  # search for result
            downloaded_files = Fido.fetch(results, path=save_path)  # download result
            result = downloaded_files  # return result
        except BaseException as e:
            print(e)
        return result

    def download_one_by_time_range(self, timestart, timeend, harpnum, timedelta=96, email=None, save_path=None):
        """
        download by time range and delta
        :param timestart: str eg "2010-01-01T00:00:00.000"
        :param timeend:  str eg "2022-01-01T00:00:00.000"
        :param harpnum: str eg "235"
        :param timedelta: int eg 96
        :param email: str eg "eg@eg.com"
        :param save_path: str path which to save file
        :return: result set or False
        """
        result = False
        try:
            if email is None:
                if isinstance(self.mail_address_list, list) and len(self.mail_address_list) > 0:
                    email = self.mail_address_list[0]
            if save_path is None:
                if os.path.exists(self.data_save_root_path):
                    save_path = self.data_save_root_path
            search_tuple = (
                a.Time(timestart, timeend),
                # set time range, if want use delta must set this ("2010-01-01T00:00:00.000", "2022-01-01T00:00:00.000"),

                a.jsoc.Series("hmi.sharp_cea_720s"),
                # choose series  http://jsoc.stanford.edu/JsocSeries_DataProducts_map.html

                a.jsoc.Notify(email),  # a.jsoc.Notify("eg@gmail.com"),

                a.jsoc.Segment("Bp"),  # choose segment
                a.jsoc.Segment("Bt"),
                a.jsoc.Segment("Br"),

                a.jsoc.PrimeKey("HARPNUM", str(harpnum)),

                # a.jsoc.Keyword("LON_MIN") > -70, # JSOC keyword filtering with Fido  https://docs.sunpy.org/en/stable/whatsnew/3.1.html?highlight=t_rec#jsoc-keyword-filtering-with-fido
                # a.jsoc.Keyword("LON_MAX") < 70,

                a.Sample(timedelta * u.min),  # eg a.Sample(96 * u.min) time delta, need set time range before
            )

            results = Fido.search(*search_tuple)
            downloaded_files = Fido.fetch(results, path=save_path)
            result = downloaded_files
        except BaseException as e:
            print(e)
        return result

    def __download_one_by_time_point_concurrent(self, timepoint, harpnum, job_num=0, save_path=None):
        # harpnum = harpnum  # "7302"
        # stime=timepoint # "2018-09-08T14:24:00.000"
        print('job num: {} / pid : {} is runing\n'.format(job_num, os.getpid()))
        email_len = len(self.mail_address_list)
        mail_num = job_num % email_len
        result = self.download_one_by_time_point(timepoint, harpnum, self.mail_address_list[mail_num], save_path)
        return result

    def download_some_by_time_point(self, timepoint_list, harpnum_list):
        """

        :param timepoint_list:
        :param harpnum_list:
        :return:
        """
        result = []
        t_len = len(timepoint_list)
        h_len = len(harpnum_list)
        if t_len == h_len:
            for i in range(t_len):
                this_result = self.download_one_by_time_point(timepoint_list[i], harpnum_list[i])
                result.append(this_result)
        return result

    def download_some_by_time_point_concurrent(self, timepoint_list, harpnum_list):
        """
        Give a list of times and numbers to download data in parallel,Note that the two dimensions should be consistent
        :param timepoint_list: eg ["2018-09-08T14:24:00.000","2018-09-08T00:00:00.000"]
        :param harpnum_list: eg ["7020","7020"]
        :return: None
        """
        result = []
        t_len = len(timepoint_list)
        h_len = len(harpnum_list)
        if t_len == h_len:
            max_workers = len(self.mail_address_list)
            executor = ProcessPoolExecutor(max_workers=max_workers)
            for i in range(t_len):
                future = executor.submit(
                    self.__download_one_by_time_point_concurrent,
                    timepoint_list[i],
                    harpnum_list[i],
                    i)
                result.append(future)
            executor.shutdown(True)

    def tran_json_file_tai_num_time_to_download_format(self, raw_str):
        """
        Convert the file time downloaded by json to the parameter time required for download,
        eg "7020.20170524_222400_TAI" to ["7020","2018-09-08T14:24:00.000"]
        Both in the list are in str format
        :param raw_str: str, eg "7020.20170524_222400_TAI"
        :return:  [ str, str ] -> [ harpnum , timestr ], eg ["7020","2018-09-08T14:24:00.000"],
        If the over-feed value is malformed, the conversion fails and is returned None，
        """
        result = None
        if isinstance(raw_str, str):
            str_list = raw_str.split(".")  # split by . to ["7020","20170524_222400_TAI"]
            if len(str_list) == 2:
                if len(str_list[1]) >= 15:
                    try:
                        hnum = int(str_list[0])  # try if list[1] is not int
                        this_list = str_list[1]
                        stime = "20{}-{}-{}T{}:{}:00.000".format(
                            this_list[2:4],
                            this_list[4:6],
                            this_list[6:8],
                            this_list[9:11],
                            this_list[11:13],
                        )  # "2018-09-08T14:24:00.000"
                        result = [hnum, stime]
                    except BaseException as e:
                        print(e)
        return result

    def get_job_list_from_file(self, file_path):
        """
        The file content is like "7020.20170524_000000_TAI\n7020.20170524_222400_TAI\n"
        :param file_path: str or os.path format
        :return:  ["7020.20170524_000000_TAI","7020.20170524_222400_TAI"]
        """
        result = []
        if os.path.exists(file_path):
            with open(file_path, "r") as f:
                all_list = f.read().split("\n")
            result = all_list
        return result

    def demo_download_one(self):
        self.download_one_by_time_point("2018-08-23T17:36:00.000", "7300",email="eg@eg.com",save_path="/data")

    def demo_download_some_from_file(self, file_path=None, save_path=None, mail_list=None):
        """
        get work job from file and run concurrent
        The file content is like "7020.20170524_000000_TAI\n7020.20170524_222400_TAI\n"
        :param file_path: str or os.path format
        :param save_path: the path to save file
        :param mail_list: ["eg@eg.com","eg2@eg2.com"] which has registered in jsoc before
        :return:  None
        """
        if file_path is None:
            return False
        if file_path is not None:
            self.data_save_root_path = save_path
        if mail_list is not None:
            self.mail_address_list = mail_list

        job_raw_list = self.get_job_list_from_file(file_path)
        num_list = []
        time_list = []
        for job_raw in job_raw_list:
            job_tran = self.tran_json_file_tai_num_time_to_download_format(job_raw)
            if job_tran is not None:
                num_list.append(job_tran[0])
                time_list.append(job_tran[1])
        self.download_some_by_time_point_concurrent(time_list, num_list)



if __name__ == "__main__":
    print("start run")

    d = DownloadJsoc()
    file_path = r"/www/wwwroot/app_run/data/a"
    save_path = r"/www/wwwroot/app_run/data"
    mail_list = ["xxx@foxmail.com",
                 "xxx@outlook.com",
                 "xxxx@outlook.com",
                 "xxxxx@outlook.com", ]
    d.demo_download_some_from_file(file_path, save_path, mail_list)






# coding=utf-8

"""
Purpose:   [1] plot the nlfff product (hdf format) to picture

Usage:     This code depends on the numpy h5py matplotlib
           The first librarie is in the standard Anaconda distribution.
           The h5py matplotlib library can be obtained from pip
           This code is compatible with python 3.7.x.

Examples:  None Now

Adapted:   ZhaoZhongRui (zhaozhongrui21@mails.ucas.ac.cn) Edit Python (2022.03)

"""

import numpy as np
import matplotlib.pyplot as plt
import h5py
import time


class NlfffPlotD3CutCake():
    """
    Draw a portion slice, like cut a cake
    """

    def __init__(self, *args, **kwargs):
        """
        init this function,you can set some values for matplotlib.pyplot.figure
        eg: d3_drawer=NlfffPlotD3CutCake(figsize=(6,9))
        :param args:
        :param kwargs:
        """

        # set colormap
        self.color_map = None # eg plt.get_cmap('coolwarm')
        self.colormap_set_max = 0  # colormap will show range [colormap_max,colormap_min]
        self.colormap_set_min = 0
        self.colormap_alpha = "auto" # None or auto or float, eg 0.8  set for global transparency
        self.colormap_out_range_display=True #True #False #True  # set True , Values that exceed the range are compressed to the maximum value instead of not being displayed
        self.colormap_auto_value = True  # auto change colormap_max,colormap_min by real value
        self.colormap_auto_zip = 0.8  # range (0 - 1]  for zip [colormap_max,colormap_min] , and 1 is not zip
        self.colormap_auto_mirror = True # If set True, and the maximum minimum value is positive and negative, respectively, then the final value becomes about 0 symmetry
        self.colorbar_show = True  # show colorbar in picture
        self.cut_line_edges = dict(color='0.6', linewidth=0.6, zorder=1e3)  # if you set None,it will not plot the lines
        self.cut_line_cross = dict(color='0.4', linewidth=1, zorder=1e3)

        # inner use
        self.__cut_list = []  # for add cut job
        self.__hdf_path = None  # for add path
        self.full_data_array = None
        self.set_default_cmap()


    def set_default_cmap(self):
        lut=8
        if self.color_map is None:
            self.color_map=plt.get_cmap('coolwarm',lut=lut)
        if self.colormap_alpha == "auto":
            a = np.append(np.linspace(0, 1, int(lut / 2)), np.linspace(1, 0, int(lut / 2)))
            self.color_map.set_gamma(a)


    def load_data_array(self, array_data):
        """
        load data from array
        :param array_data: numpy.ndarray shape like (3,Nx,Ny,Nz)
        :return:load sucessfully return True, error type or shape return False
        """
        result = False
        if isinstance(array_data, np.ndarray):
            if len(array_data.shape) == 4:
                self.full_data_array = array_data
                result = True
        return result

    def load_data_hdf(self, hdf_path):
        """
        given the path of hdf, it will load the data inner for next step use
        :param hdf_path: str or os.path
        :return: open sucessfully return True else return False
        """
        if self.__hdf_path is not None:
            self.close_data_hdf()
        try:
            self.__hdf_path = hdf_path
            self.__hf = h5py.File(self.__hdf_path, 'r')
            self.full_data_array = self.__hf[self.data_hdf_dataset_name]  # four D
            # print(self.full_data_array[0]) # three D
            # print(type(self.full_data_array[0])) #<class 'numpy.ndarray'>
            result = True
        except BaseException as e:
            print(e)
            result = False
        return result

    def close_data_hdf(self):
        """
        close the hdf file which haven open
        :return:
        """
        result = False
        try:
            self.__hf.close()
            result = True
        except BaseException as e:
            print(e)
        # time.sleep(100)
        return result

    def add_cut(self, B="Bx", N="Nx", cut_num=None, cut_percent=0.5):
        """
        when you haven load hdf data,
        add cut surface for cut draw, which can be use many times for different surfance in one picture
        :param B: 'Bx' 'By' or 'Bz'; choose one Magnetic field component to draw
        :param N: 'Nx' 'Ny' or 'Nz'; choose cut in which angle; when choose 'Nz', The slice will be perpendicular to z axis
        :param cut_num: int type , set for cut where, can be [0, max(axis)], if not set this,will use cut_percent default
        :param cut_percent: default 0.5, percent of axis
        :return: None
        """
        this_cut_dict = dict(B=B, N=N, cut_percent=cut_percent, cut_num=cut_num)
        self.__cut_list.append(this_cut_dict)

    def run_cut(self,
                fig_dict={"figsize":(6, 6)},

                # https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.show.html
                is_show=True,
                show_dict={"block":True},

                # https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.savefig.html
                is_save=False,
                save_dict={"fname":"./cut.png", "dpi":'figure'}
                ):
        """
        when you haven load data and add_cut ,
        run cut for computer data and prepare for draw
        :return: None
        """
        # computer shape
                # set fig and you can use it outside like : d3_drawer.fig.XXX
        self.fig = plt.figure(**fig_dict)

        # set draw view and you can use it outside like : d3_drawer.ax.view_init(40, -30)
        self.ax = self.fig.add_subplot(projection='3d')
        ## Set distance and angle view
        self.ax.view_init(-40, 30)
        self.ax.dist = 11
        self.data_hdf_dataset_name="Bxyz"
        ## Set labels and zticks
        self.ax.set(
            xlabel='X',
            ylabel='Y',
            zlabel='Z',
        )

        data_shape = self.full_data_array[0].shape
        self.__Nx = data_shape[1]
        self.__Ny = data_shape[0]
        self.__Nz = data_shape[2]
        self.__X, self.__Y, self.__Z = np.meshgrid(np.arange(self.__Nx), np.arange(self.__Ny), np.arange(self.__Nz))

        # prepare job
        contourf_list = []  # for draw dict
        x_value_list = []  # for cut line
        y_value_list = []
        z_value_list = []
        for cut in self.__cut_list:
            if cut["B"] == "Bx":
                bxyz_num = 0
            elif cut["B"] == "By":
                bxyz_num = 1
            elif cut["B"] == "Bz":
                bxyz_num = 2
            else:
                break
            if cut["N"] == "Nx":
                if cut["cut_num"] is not None and 0 <= cut["cut_num"] <= self.__Nx:  # set direct value
                    x_value = int(cut["cut_num"])
                else:
                    x_value = int(self.__Nx * cut["cut_percent"])  # computer value by percent
                data1_x = self.full_data_array[bxyz_num][:, x_value, :]
                if self.colormap_auto_value:  # set auto for computer max and min
                    self.colormap_set_max = max(self.colormap_set_max, data1_x.max())
                    self.colormap_set_min = min(self.colormap_set_min, data1_x.min())
                data2_y = self.__Y[:, x_value, :]
                data3_z = self.__Z[:, x_value, :]
                draw_job_dict = dict(X=data1_x, Y=data2_y, Z=data3_z, zdir="x", offset=x_value)
                contourf_list.append(draw_job_dict)
                x_value_list.append(x_value)  # for draw cut line
            elif cut["N"] == "Ny":
                if cut["cut_num"] is not None and 0 <= cut["cut_num"] <= self.__Ny:
                    y_value = int(cut["cut_num"])
                else:
                    y_value = int(self.__Ny * cut["cut_percent"])
                data1_x = self.__X[y_value, :, :]
                data2_y = self.full_data_array[bxyz_num][y_value, :, :]
                if self.colormap_auto_value:
                    self.colormap_set_max = max(self.colormap_set_max, data2_y.max())
                    self.colormap_set_min = min(self.colormap_set_min, data2_y.min())
                data3_z = self.__Z[y_value, :, :]
                draw_job_dict = dict(X=data1_x, Y=data2_y, Z=data3_z, zdir="y", offset=y_value)
                contourf_list.append(draw_job_dict)
                y_value_list.append(y_value)
            elif cut["N"] == "Nz":
                if cut["cut_num"] is not None and 0 <= cut["cut_num"] <= self.__Nz:
                    z_value = int(cut["cut_num"])
                else:
                    z_value = int(self.__Nz * cut["cut_percent"])
                data1_x = self.__X[:, :, z_value]
                data2_y = self.__Y[:, :, z_value]
                data3_z = self.full_data_array[bxyz_num][:, :, z_value]
                if self.colormap_auto_value:
                    self.colormap_set_max = max(self.colormap_set_max, data3_z.max())
                    self.colormap_set_min = min(self.colormap_set_min, data3_z.min())
                draw_job_dict = dict(X=data1_x, Y=data2_y, Z=data3_z, zdir="z", offset=z_value)
                contourf_list.append(draw_job_dict)
                z_value_list.append(z_value)
        if 1 > self.colormap_auto_zip > 0 and self.colormap_auto_value:  # for zip color map
            c_min = self.colormap_set_min * self.colormap_auto_zip
            c_max = self.colormap_set_max * self.colormap_auto_zip
        else:
            c_min = self.colormap_set_min
            c_max = self.colormap_set_max

        if self.colormap_auto_mirror and c_min*c_max < 0: # set True, and the maximum minimum value is positive and negative
            max_abs=max(abs(c_max),abs(c_min))
            c_max=max_abs
            c_min=-max_abs

        kw = {
            'vmin': c_min,
            'vmax': c_max,
            # 'levels': np.linspace(c_min, c_max, 10),
            'cmap': self.color_map,
            # "c": np.linspace(c_min, c_max, 10) / 2550,

            # 'alpha': self.colormap_alpha,
        }
        if c_min != c_max: # if same will raise error
            kw["levels"]=np.linspace(c_min, c_max, 19)
        if isinstance(self.colormap_alpha,float) and 0<self.colormap_alpha <1:
            kw["alpha"]=self.colormap_alpha

        # draw
        cut_obj_list = []  # cut obj  for band colorbar and other use
        for cut_job in contourf_list:
            if self.colormap_out_range_display:
                if cut_job["zdir"] == "x":
                    cut_job["X"] = np.clip(cut_job["X"], a_min=c_min, a_max=c_max)
                elif cut_job["zdir"] == "y":
                    cut_job["Y"] = np.clip(cut_job["Y"], a_min=c_min, a_max=c_max)
                elif cut_job["zdir"] == "z":
                    cut_job["Z"] = np.clip(cut_job["Z"], a_min=c_min, a_max=c_max)
            cut_job.update(kw)  # add common dict
            cut_obj = self.ax.contourf(**cut_job)
            cut_obj_list.append(cut_obj)

        # set axis
        xmin, xmax = 0, self.__Nx
        ymin, ymax = 0, self.__Ny
        zmin, zmax = 0, self.__Nz
        self.ax.set(xlim=[xmin, xmax], ylim=[ymin, ymax], zlim=[zmin, zmax])

        # draw line
        x_value_len = len(x_value_list)
        y_value_len = len(y_value_list)
        z_value_len = len(z_value_list)
        # for cut_line_cross
        if isinstance(self.cut_line_cross, dict):
            edges_kw = self.cut_line_cross
            ax = self.ax
            if y_value_len and z_value_len:  # if len is 0 , it is False
                for y_value in y_value_list:
                    for z_value in z_value_list:
                        ax.plot([xmin, xmax], [y_value, y_value], [z_value, z_value], **edges_kw)
            if x_value_len and z_value_len:
                for x_value in x_value_list:
                    for z_value in z_value_list:
                        ax.plot([x_value, x_value], [ymin, ymax], [z_value, z_value], **edges_kw)
            if x_value_len and y_value_len:
                for x_value in x_value_list:
                    for y_value in y_value_list:
                        ax.plot([x_value, x_value], [y_value, y_value], [zmin, zmax], **edges_kw)
        # for cut_line_edges
        if isinstance(self.cut_line_edges, dict):
            edges_kw = self.cut_line_edges
            ax = self.ax
            if x_value_len:
                for x_value in x_value_list:
                    ax.plot([x_value, x_value], [ymin, ymin], [zmin, zmax], **edges_kw)
                    ax.plot([x_value, x_value], [ymax, ymax], [zmin, zmax], **edges_kw)
                    ax.plot([x_value, x_value], [ymin, ymax], [zmin, zmin], **edges_kw)
                    ax.plot([x_value, x_value], [ymin, ymax], [zmax, zmax], **edges_kw)
            if y_value_len:
                for y_value in y_value_list:
                    ax.plot([xmin, xmax], [y_value, y_value], [zmin, zmin], **edges_kw)
                    ax.plot([xmin, xmax], [y_value, y_value], [zmax, zmax], **edges_kw)
                    ax.plot([xmin, xmin], [y_value, y_value], [zmin, zmax], **edges_kw)
                    ax.plot([xmax, xmax], [y_value, y_value], [zmin, zmax], **edges_kw)
            if z_value_len:
                for z_value in z_value_list:
                    ax.plot([xmin, xmin], [ymin, ymax], [z_value, z_value], **edges_kw)
                    ax.plot([xmax, xmax], [ymin, ymax], [z_value, z_value], **edges_kw)
                    ax.plot([xmin, xmax], [ymin, ymin], [z_value, z_value], **edges_kw)
                    ax.plot([xmin, xmax], [ymax, ymax], [z_value, z_value], **edges_kw)

        # add color bar
        if len(cut_obj_list) > 0 and self.colorbar_show:
            self.fig.colorbar(cut_obj_list[0], ax=self.ax, fraction=0.02, pad=0.1)  # , label='Name [units]')
        if is_save:
            plt.savefig(**save_dict)
        if is_show:
            plt.show(**show_dict)
        



import pickle
import numpy as np
import datetime
import wget
import pickle
import os
from peewee import *
import time


class HarpNoaa():
    def __init__(self):
        self.all_harps_with_noaa_ars_online="http://jsoc.stanford.edu/doc/data/hmi/harpnum_to_noaa/all_harps_with_noaa_ars.txt"
        self.all_harps_with_noaa_ars_local="all_harps_with_noaa_ars.txt"
        self.kharp_vnoaalist_dict=dict()
        self.knoaa_vharp_dict=dict()
        self.noaa_list=list()
        self.harp_list=list()
        self.has_prepare=False

    
    def prepare(self):
        if not os.path.exists(self.all_harps_with_noaa_ars_local):
            wget.download(self.all_harps_with_noaa_ars_online,out=self.all_harps_with_noaa_ars_local)
        with open(self.all_harps_with_noaa_ars_local,"r")as f:
            all=f.read().split("\n")
        for i in all:
            try:
                harp_num=str(int(i.split(" ")[0]))
                noaa_list=i.split(" ")[1].split(",")
                self.kharp_vnoaalist_dict[harp_num]=noaa_list
                for noaa_num in noaa_list:
                    self.knoaa_vharp_dict[noaa_num]=harp_num
            except BaseException as e:
                print(i)
                print(e)
        self.noaa_list=list(self.knoaa_vharp_dict)
        self.harp_list=list(self.kharp_vnoaalist_dict)
        self.has_prepare=True


    def save(self,kharp_vnoaalist_dict_pickle_path=None,knoaa_vharp_dict_pickle_path=None,noaa_list_path=None,harp_list_path=None):
        if not self.has_prepare:
            self.prepare()
        if kharp_vnoaalist_dict_pickle_path is not None:
            with open(kharp_vnoaalist_dict_pickle_path,'wb+') as pickle_file:  
                pickle.dump(self.kharp_vnoaalist_dict,pickle_file)          
        if knoaa_vharp_dict_pickle_path is not None:
            with open(knoaa_vharp_dict_pickle_path,'wb+') as pickle_file: 
                pickle.dump(self.knoaa_vharp_dict,pickle_file) 
        if noaa_list_path is not None:
            with open(noaa_list_path,"w+") as f:
                f.write("\n".join(self.noaa_list))
        if harp_list_path is not None:
            with open(harp_list_path,"w+") as f:
                f.write("\n".join(self.harp_list))
                



class Flare(Model):
    deeps_flare_id = IntegerField(null=True)
    latitude = IntegerField(null=True)
    longtitude = IntegerField(null=True)
    noaa_ar = IntegerField(null=True)
    end_datetime = DateTimeField(null=True)
    peak_datetime = DateTimeField(null=True)
    source = CharField(null=True)
    start_datetime = DateTimeField(null=True)
    xray_class = CharField(null=True)
    xray_intensity = IntegerField(null=True)
    # id xray_class xray_intensity latitude longtitude start_datetime peak_datetime end_datetime source
    class Meta:
        database = None
        table_name = 'a202203_flare'
        primary_key = False
    

class NoaaFlaretime():
    def __init__(self,db_name,config_dict):
        # db_name='deepsolar'
        # config_dict={'host': '10.10.1.1', 'port': 1234, 'user': 'deepsolar', 'password': 'xxxxxxxx'}
        self.Flare=Flare
        self.Flare._meta.database=PostgresqlDatabase(db_name, **config_dict)
        self.noaa_list=[]
        self.noaa_flaretime_dict=dict()


    def set_noaa_list(self,noaa_list=None,noaa_list_path=None):
        if noaa_list is not None:
            self.noaa_list=noaa_list
        if noaa_list_path is not None:
            with open(noaa_list_path) as f:
                allstrraw=f.read()
            self.noaa_list=allstrraw.split("\n")
            
    def deal_one_noaa_num(self,noaa_num,save_path=None):
        noaa_num=int(noaa_num)
        if save_path is not None and not os.path.exists(save_path):
            os.makedirs(save_path)
            is_save=True
        else:
            is_save=False
        all=[[0,0,0,-1]]
        for i in self.Flare.select().where(self.Flare.noaa_ar==noaa_num):
            # # print(type(i),i.id,i.xray_class,i.xray_intensity)
            # # print(type(i.start_datetime),i.end_datetime,int(i.end_datetime.timestamp()))
            if i.start_datetime is  None and i.end_datetime is  None:
                continue
            end_t=int(i.end_datetime.timestamp())
            start_t=int(i.start_datetime.timestamp())
            map_dict={"A":100,"B":200,"C":300,"M":400,"X":500}  
            if i.xray_class not in map_dict.keys():
                continue
            if isinstance( i.xray_intensity,int):
                sub_v=i.xray_intensity
            else:
                sub_v=0
            xray_value=map_dict[i.xray_class]+sub_v
            all.append([start_t,end_t,xray_value,i.deeps_flare_id])
        if len(all)==1:
            return False
        a=np.array(all)
        
        s=a[:,0]
        e=a[:,1]
        l=a[:,2]
        i=a[:,3]
        new_order=list(np.lexsort((l,s))) 

        a=a[new_order,:]
        a[0,0]=a[1,0]
        a[0,1]=a[-1,1]
        a=a.astype(int)
        flare_delta_num=a.shape[0]

        if is_save:
            this_p=os.path.join(save_path,"{}.f0.csv".format(noaa_num))
            np.savetxt(this_p,a,fmt='%i', delimiter=",")
        b=a
        
        i_have_add=0
        if flare_delta_num>1:
            for fi in range(1,flare_delta_num-1):
                if a[fi,1]<a[fi+1,0]:
                    this_insert_data=np.array([a[fi,1]+1,a[fi+1,0]-1,0,-1])
                    b=np.insert(b,fi+i_have_add+1,this_insert_data,axis=0)
                    i_have_add+=1
                elif a[fi,1]>a[fi+1,0]:
                    b[fi+i_have_add,1]=a[fi+1,0]-1
            
        if is_save:
            this_p=os.path.join(save_path,"{}.f1.csv".format(noaa_num))
            np.savetxt(this_p,a,fmt='%i', delimiter=",")
            # # print(a)
            this_p=os.path.join(save_path,"{}.f2.csv".format(noaa_num))
            np.savetxt(this_p,b,fmt='%i', delimiter=",")
            # # print(b)
        return b
        
    def deal_noaa_num(self,save_split_numpy_path=None):
        for i in self.noaa_list:
            print(i)
            try:
                this_r=self.deal_one_noaa_num(i,save_split_numpy_path)
                if this_r is not False:
                    self.noaa_flaretime_dict[str(i)]=this_r
                else:
                    print("{}.none".format(i))
            except BaseException as e:
                print("{}.error.{}".format(i,e))
        return self.noaa_flaretime_dict

    def save_pickle(self,pickle_path):
        with open(pickle_path,'wb+') as pickle_file:  
            pickle.dump(self.noaa_flaretime_dict,pickle_file)  
        
    def load_pickle(self,pickle_path):
        with  open(pickle_path,'rb') as pickle_file:   
            self.noaa_flaretime_dict=pickle.load(pickle_file) 
        



class DealFlareLevel():
    def __init__(self,noaa_flaretime_dict_pickle_path):
        # noaa_dict_path='/home/zzr/project2/tag/noaa_flare_time_dict.pickle'
        with open(noaa_flaretime_dict_pickle_path,'rb')  as f:
            self.noaa_dict = pickle.load(f)
            
    def get_one_noaa_time_in_level(self,noaa_ar,data_time,time_type="%Y%m%d_%H%M%S"):
        # data_time="20150303_204800"
        # noaa_ar=12292
        
        noaa_dict=self.noaa_dict
        if str(noaa_ar) not in noaa_dict.keys():
            return False
        
        this_time=int(datetime.datetime.strptime(data_time,time_type).timestamp())
        # this_time
        this_noaa_value=noaa_dict[str(noaa_ar)]
        raw_t=this_noaa_value[:,0:2]
        this_t=np.ones_like(raw_t)*this_time
        time_delta=this_t-raw_t
        flag=np.sign(time_delta[:,0])*np.sign(time_delta[:,1])<=0
        time_is_in=np.where(flag)[0]
        if len(time_is_in)>0:
            time_is_index=time_is_in[-1]
            this_value=this_noaa_value[time_is_index,:]
            # # # # print(this_value)
            this_level=this_value[2]
            this_id=this_value[-1]
            # # # # print(this_level,this_id)
        else:
            this_level=0
            this_id=-1
        result=[this_level,this_id]
        return result
        
        

        # noaa_dict=noaa_dict
    def get_one_noaa_timelist_maxlevel(self,noaa_ar,data_time,time_list_second=[6*60*60,12*60*60,24*60*60,48*60*60],time_type="%Y%m%d_%H%M%S"):
        # data_time="20150303_204800"
        # # data_time="20150303_204800"
        # # noaa_dict=noaa_dict TODO
        # noaa_ar=12292
        # time_later_list=[6*60*60,12*60*60,24*60*60,48*60*60]
        time_later_list=time_list_second
        noaa_dict=self.noaa_dict
        
        if str(noaa_ar) not in noaa_dict.keys():
            # # # print("not in")
            return False
        # this_t_delta=time_later_list[4]
        result=[]
        for this_t_delta in time_later_list:
            # # # print(this_t_delta)

            this_time=int(datetime.datetime.strptime(data_time,time_type).timestamp())
            next_time=this_time+this_t_delta
            # this_time
            this_noaa_value=noaa_dict[str(noaa_ar)]

            raw_t=this_noaa_value[:,0:2]
            this_t=np.ones_like(raw_t)*this_time
            next_t=np.ones_like(raw_t)*next_time

            this_time_delta=this_t-raw_t
            this_flag=np.sign(this_time_delta[:,0])*np.sign(this_time_delta[:,1])<=0
            # # # print(this_flag)
            this_time_is_in=np.where(this_flag)[0]
            next_time_delta=next_t-raw_t
            next_flag=np.sign(next_time_delta[:,0])*np.sign(next_time_delta[:,1])<=0
            # # # print(next_flag)
            next_time_is_in=np.where(next_flag)[0]


            if len(this_time_is_in)>0 and len(next_time_is_in)>0:
                # # # print(1)
                # # # print(this_time_is_in[-1],next_time_is_in[-1]+1)
                max_value_array=this_noaa_value[this_time_is_in[-1]:next_time_is_in[-1]+1,:]
            elif len(this_time_is_in)>0 :
                # # # print(2)
                max_value_array=this_noaa_value[this_time_is_in[-1]:,:]
            elif len(next_time_is_in)>0:
                # # # print(3)
                max_value_array=this_noaa_value[:next_time_is_in[-1]+1,:]
            elif np.sign(this_time_delta[0,0])*np.sign(this_time_delta[-1,1])<0:
                # # # print(4)
                max_value_array=this_noaa_value[:,:]
            else:
                # # # print(5)
                max_value_array=False
            # # # print(max_value_array)
                
            if max_value_array is not False:
                # # # print(len(max_value_array))
                if len(max_value_array)==1:
                    max_level=max_value_array[0,2]
                    max_id=max_value_array[0,3]
                elif len(max_value_array)>1:
                    max_index=np.argmax(max_value_array[:,2])
                    max_level=max_value_array[max_index,2]
                    max_id=max_value_array[max_index,3]
                # # # print(this_level,this_id)
                else:
                    max_level=0
                    max_id=-1
            else:
                max_level=0
                max_id=-1
            this_result=[max_level,max_id]
            result.append(this_result)
        # # # print(result)
        return result


    def deal_one_noaa_with_time(
        self,
        noaa_ar,
        data_time,
        time_list_second=[6*60*60,12*60*60,24*60*60,48*60*60],
        time_type="%Y%m%d_%H%M%S"):
        
        this_level=self.get_one_noaa_time_in_level(noaa_ar,data_time)
        range_level=self.get_one_noaa_timelist_maxlevel(noaa_ar,data_time,time_list_second,time_type)
        result=(this_level,range_level)
        return result
    
    def deal_one_or_group_noaa_with_time(
        self,
        noaa_ars,
        data_time,
        time_list_second=[6*60*60,12*60*60,24*60*60,48*60*60],
        time_type="%Y%m%d_%H%M%S"):
        """c1=w.deal_one_or_group_noaa_with_time([11147,11149],"20110122_081200")
           c1=w.deal_one_or_group_noaa_with_time([11149],"20110122_081200")
        Returns:
            tuple:   ([324, 65525], [[324, 65525], [324, 65525], [324, 65525], [324, 65525]])
                    [now flare level, now flare id] [given flare level, given flare id] 
        """
        t_level_result=[]
        t_range_result=[]
        if isinstance(noaa_ars,int) or isinstance(noaa_ars,str):
            result=self.deal_one_noaa_with_time(noaa_ars,data_time,time_list_second,time_type)
        elif isinstance(noaa_ars,list):  
            for noaa_ar in noaa_ars:
                # this_level=get_one_noaa_time_in_level(noaa_ar,data_time)
                # range_level=get_one_noaa_timelist_maxlevel(noaa_ar,data_time,time_list_second,time_type)
                # one_result=(this_level,range_level)
                # temp_result.append(one_result)
                # try:
                this_level,range_level=self.deal_one_noaa_with_time(noaa_ar,data_time,time_list_second,time_type)
                # # # print(this_level ,"00==========", range_level)
                if this_level and range_level:
                    t_level_result.append(this_level)
                    t_range_result.append(range_level)
            # print(t_range_result)
            level_result=np.array(t_level_result)
            # print(level_result)
            if len(level_result) > 0:
                # # print(level_result)
                level_index=level_result.argmax(axis=0)[1]
                # # print(level_index)
                level_result=t_level_result[level_index]
                
                # # print(t_range_result)

                max_index=np.array(t_range_result)[:,:,0]#[1]#.argmax()
                # print(max_index)
                # print(max_index.argmax(axis=0))
                max_list=max_index.argmax(axis=0) #list(max_index[:,1])
                finall_result_range=[]
                for i in range(len(max_list)):
                    # # print(i)
                    finall_result_range.append(t_range_result[max_list[i]][i])
                result=level_result,finall_result_range
                # # # print(result)
                #     return result
            else:
                result=False
        else:
            result=False
        return result







class DoSharp():
    def __init__(self,noaa_flaretime_dict_pickle_path,kharp_vnoaalist_dict_pickle_path,save_path=None):

        # noaa_flaretime_dict_pickle_path='/home/zzr/project2/tag/all_noaa_flaretime_dick_v3.pickle'
        # kharp_vnoaalist_dict_pickle_path='/home/zzr/project2/tag/noaa_to_harp.pickle'
        self.save_path=save_path #"/home/zzr/project2/tag/to_db_v3.2"
        self.dealfl=DealFlareLevel(noaa_flaretime_dict_pickle_path=noaa_flaretime_dict_pickle_path)
        with open(kharp_vnoaalist_dict_pickle_path,'rb')  as f:
            self.noaa_sharp_dict = pickle.load(f)
        self.noaa_sharp_dict_keys=list(self.noaa_sharp_dict.keys())

        
    
    def get_noaalist_from_sharpnum(self,sharp_num):
        sharp_num=str(sharp_num)
        # print(str(sharp_num)) 
        # print(self.noaa_sharp_dict.keys())
        # print(self.noaa_sharp_dict_keys)
        # print(str(sharp_num) in self.noaa_sharp_dict_keys)
        if str(sharp_num) in self.noaa_sharp_dict_keys:
            return self.noaa_sharp_dict[str(sharp_num)]
        else:
            return False
    
    # 2325.20121227_114800
    def deal_one_job(self,sharp_time,job_id=0):
        issharp=0
        iscea=1
        sharp_num,data_time=sharp_time.split(".")
        noaa_list=self.get_noaalist_from_sharpnum(sharp_num)
        # print(noaa_list)
        js=os.path.join(self.save_path,"{}.sucess.csv".format(job_id))
        jf=os.path.join(self.save_path,"{}.fail1.txt".format(job_id))
        jf2=os.path.join(self.save_path,"{}.fail2.txt".format(job_id))
        if noaa_list is not False:
            dresult=self.dealfl.deal_one_or_group_noaa_with_time(noaa_list,data_time)
            # ([0, -1], [[0, -1], [0, -1], [328, 74259], [328, 74259]])
            # print(dresult)
            if dresult is False:
                with open(jf2,"a+") as f:
                    f.write(sharp_time+"\n")   
                result = False
            else:
                sresult="{},{},{},{},{},{},{}".format(
                    sharp_time,
                    sharp_num,
                    data_time,
                    issharp,
                    iscea,
                    dresult[0][0],
                    dresult[0][1]
                )
                for i in dresult[1]:
                    sresult+=",{},{}".format(i[0],i[1])
                sresult+="\n"
                with open(js,"a+") as f:
                    f.write(sresult)
                result=True
        else:
            with open(jf,"a+") as f:
                f.write(sharp_time+"\n")     
            result=False
        return result      
    
    
    def run(self,harpnum_time_list_path,save_path=None):
        if save_path is not None:
            self.save_path=save_path
        if self.save_path is None:
            return False
        if os.path.exists(self.save_path):
            os.rename(self.save_path, "{}.bak.{}".format(self.save_path,time.time()))
        os.makedirs(self.save_path)

        #"/home/zzr/project2/tag/sharp12.txt"  # 5946.20150911_101200  list
        with open(harpnum_time_list_path,"r") as f:
            jobs=f.read().split("\n")
        lens=len(jobs)
        i=0
        for job in jobs:
            try:
                result=self.deal_one_job(job,0)
                if result is False:
                    js=os.path.join(self.save_path,"{}.runfail.txt".format(0))
                    with open(js,"a+") as f:
                        f.write(job+"\n")
            except BaseException as e:
                print(job)
                print(e)
                js=os.path.join(self.save_path,"{}.cantrun.txt".format(0))
                with open(js,"a+") as f:
                    f.write(job+"\n")
            i+=1
            if i%1000==0:
                print("{}/{}, {}".format(i,lens,i/lens))




import os
import re


def write_set_to_path(raw_set,raw_path,need_root=False):
    for i in raw_set:
        if not need_root:
            i=i.split("/")[-1]
        with open(raw_path,"a+") as f:
            f.write(str(i)+"\n")


def check_quality_from_log(file_path):
    """
    Detect whether the file degree angle is greater than 30,
    greater than 30 or the file is empty, return False,
    otherwise True
    :param file_path: NLFFFquality3.log path
    :return:
    """
    result = False
    if os.path.exists(file_path):
        with open(file_path, 'r')as f:
            file_data = f.read()
        # print(file_data)
        pattern = r'Angle.*?Degree'
        match = re.findall(pattern, file_data)
        # print(match)
        p = r'\d+\.?\d+'
        nums = re.findall(p, str(match))
        if len(nums) == 0:
            result = False
        else:
            result = True
            for num in nums:
                if float(num) > 30.0:
                    result_n = False
                else:
                    result_n = True
                result = result and result_n  # ensure every value is ok
    return result





def clean_to_one_step(raw_path):
    init_set=set(['allboundaries3.dat', 'allboundaries1.dat', 'grid3.ini', 'boundary.ini', 'mask2.dat', 'mask3.dat', 'grid2.ini', 'grid1.ini', 'mask1.dat', 'allboundaries2.dat','run.log'])
    dict_set=set(next(os.walk(raw_path))[2])
    rmf=dict_set-init_set
    result=[]
    for rmo in rmf:
        rp=os.path.join(raw_path,rmo)
        if os.path.isfile(rp):
            # print(rp)
            result.append(rp)
    return result
            
# clean_to_one("/home/ma-user/work/pro/num_1800_1899/hmi.sharp_cea_720s.1834.20120708_044800_TAI")


def run_root_path(path_root,log_root,clean_to_one=False):
    # path_root="/home/ma-user/work/pro/archive-202209/grid2.done"
    todo1=os.path.join(log_root,"todo1.txt")
    todo2=os.path.join(log_root,"todo2.txt")
    todo3=os.path.join(log_root,"todo3.txt")
    done3=os.path.join(log_root,"done3.txt")
    clean1=os.path.join(log_root,"clean1.txt")
    cleanp=os.path.join(log_root,"clean1.sh")
    all_l=os.walk(path_root)

    set_1_todo=set()
    set_2_todo=set()
    set_3_todo=set()
    clean_to_1_todo=set()
    set_3_done=set()


    for one_l in all_l:
        this_path=one_l[0]
        this_file_list=one_l[2]
        this_file_count=len(this_file_list)
        if this_file_count>10:
            # print(one_l)
            # pass
            if "NLFFFquality3.log" in this_file_list:
                if check_quality_from_log(os.path.join(this_path,"NLFFFquality3.log")):
                    set_3_done.add(this_path)
                else:
                    clean_to_1_todo.add(this_path)
            elif "NLFFFquality2.log" in this_file_list:
                if check_quality_from_log(os.path.join(this_path,"NLFFFquality2.log")):
                    if "grid.ini" in this_file_list:
                        if os.path.getmtime(os.path.join(this_path,"grid.ini"))<os.path.getmtime(os.path.join(this_path,"NLFFFquality2.log")):
                            set_3_todo.add(this_path)
                        else:
                            clean_to_1_todo.add(this_path)
                    else:
                        clean_to_1_todo.add(this_path)
                else:
                    clean_to_1_todo.add(this_path)
            elif "NLFFFquality1.log" in this_file_list:
                if check_quality_from_log(os.path.join(this_path,"NLFFFquality1.log")):
                    if "grid.ini" in this_file_list:
                        if os.path.getmtime(os.path.join(this_path,"grid.ini"))<os.path.getmtime(os.path.join(this_path,"NLFFFquality1.log")):
                            set_2_todo.add(this_path)
                        else:
                            clean_to_1_todo.add(this_path)
                    else:
                        clean_to_1_todo.add(this_path)
                else:
                    clean_to_1_todo.add(this_path) 
            else:
                clean_to_1_todo.add(this_path) 

        elif this_file_count==10 and "grid3.ini" in this_file_list:
            # print(one_l)
            set_1_todo.add(this_path)
    clean_sh=set()
    if clean_to_one:
        
        for job in clean_to_1_todo:
            c_result = clean_to_one_step(job)
            # for c in c_result:
            clean_sh= clean_sh | set(c_result)


    
    write_set_to_path(set_1_todo,todo1)
    write_set_to_path(clean_to_1_todo,clean1)
    write_set_to_path(set_2_todo,todo2)
    write_set_to_path(set_3_todo,todo3)
    write_set_to_path(set_3_done,done3,True)
    write_set_to_path(clean_sh,cleanp,True)

if __name__=="__main__":
    run_root_path(
        "/home/ma-user/work/pro/archive-202209/grid2.done",
        "/home/ma-user/work/pro/log/new",
        True
    )





import pathlib
from setuptools import setup, find_packages

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name='pynlfff',
    version='0.3.3.4',
    description='python for nlfff',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/deepsolar/pynlfff',
    author='ZhongRui Zhao, Xinze Zhang',
    author_email='zhaozhongrui21@mails.ucas.ac.cn',
    license='MIT',
    keywords='python nlfff',
    packages=find_packages(),
    install_requires=['numpy', 'h5py', 'matplotlib'],
    python_requires='>=3'
)



# coding=utf-8
import subprocess
import sys
import os
import re
import shutil
import multiprocessing
from .prepare_base import PrepareWorker


# ------全局控制------------------------------------------------
def main_control():
    print("开始运行")
    config_dict = get_config()
    if config_dict:
        print(config_dict)
        data_root = config_dict["data_root"]
        # "/home/zzr/data/sharp_upload_1/num_0001_0999/num_0001_0099/datalist.txt"
        data_job_list_path = os.path.join(data_root, "datalist.txt")
        if os.path.exists(data_job_list_path):
            job_list = find_joblist_from_datalist(data_job_list_path)
        else:
            job_list = find_joblist_from_dir(data_root)
        print("job num {}".format(len(job_list)))
        # app_root = config_dict["clinff_root"]
        save_path = config_dict["save_root"]
        work_num = config_dict["work_num"]
        # runspace_root=config_dict["runspace_root"]
        # grid_level=config_dict["grid_level"]
        run_result = manager_mult_pro(data_root, job_list, save_path,work_num)
        print("运行结果{}".format(run_result))
    else:
        print("参数或配置出现错误")
    return True


# -------外部接口--------------------------------------------
# 获取主程序运行根目录和数据存储目录并返回


def get_config():
    if len(sys.argv) < 4:
        print("请输入6个参数：")
        result = False
    else:
        # clinff_root = sys.argv[1]  # c主程序根目录
        data_root = sys.argv[1]  # 数据根目录
        save_root = sys.argv[2]  # 文件保存根目录
        work_num = sys.argv[3]  # 同时运行任务，0所有核心全部用上 (0,1)所有核心的几分之几  >=1 用设置的核心  其他为1  

        # runspace_root = sys.argv[5]  # 运行空间，如果是obs，需要先拷贝运行空间再运算，再复制最终空间
        # grid_level=sys.argv[6] # 运行层级 0预处理  123层 4质量控制
        # if os.path.exists(data_root) and os.path.exists(save_root) :
        result = dict(data_root=data_root, save_root=save_root,work_num=work_num)
        # else:
        #     print("输入路径不存在")
        #     result = False
    return result


# ------总体控制-----------------------------------------
# 从根目录找到任务文件
# /home/zzr/data/sharp_upload_1/num_0001_0999/num_0001_0099/datalist.txt
def find_joblist_from_datalist(datalist_file_path):
    with open(datalist_file_path, "r") as f:  # 设置文件对象
        list_data = f.read()  # 可以是随便对文件的操作
    # print(list_data) # 一整个字符串类型
    # print(re.split(r'Br.fits\n|Bt.fits\n|Bp.fits\n', list_data))
    three_list = re.split(r'.Br.fits\n|.Bt.fits\n|.Bp.fits\n', list_data)
    one_list = sorted(list(set(three_list)))  # 去重+排序
    print(one_list)
    job_list = one_list
    return job_list


def find_joblist_from_dir(data_root):
    file_list = os.listdir(data_root)
    new_list = list()
    for f_name in file_list:
        a1 = f_name.replace(".Bt.fits", "")
        a2 = a1.replace(".Bp.fits", "")
        a3 = a2.replace(".Br.fits", "")
        new_list.append(a3)
    one_list = sorted(list(set(new_list)))  # 去重+排序
    print(one_list)
    return one_list  # ['hmi.sharp_cea_720s.998.20111027_093600_TAI.']


# 创建多线程生成任务
def manager_mult_pro(data_root, job_list, save_path,work_num):
    # "/home/zzr/project/linff/linff-main/deploy/space/log"
    log_path = os.path.join(save_path, "todo0.log.txt")
    work_num = get_work_num(work_num)
    with open(log_path, 'a+') as f:
        job_len = len(job_list)
        line = "============================================================\n{}\n{}\nwork_num:{}\n".format(job_list,job_len,work_num)
        f.write(line)
        
    pool = multiprocessing.Pool(processes=work_num)
    for job in job_list:
        pool.apply_async(func=manager_one_job, args=(
            data_root, job,save_path), callback=setcallback_write_log)
    pool.close()
    pool.join()
    print("全部运行结束")
    return True

# 同时运行任务，0所有核心全部用上 (0,1)所有核心的几分之几  >=1 用设置的核心  其他为1 
def get_work_num(work_num):
    # from multiprocessing import cpu_count
    # print("CPU的核数为：{}".format(multiprocessing.cpu_count()))
    # print(type(cpu_count()))
    work_num=float(work_num)
    if work_num >=1:
        real_work_num=int(work_num)
    elif (0 < work_num and work_num < 1):
        all_cpu=multiprocessing.cpu_count()
        real_work_num=int(all_cpu*work_num)
    elif work_num == 0:
        all_cpu=multiprocessing.cpu_count()
        real_work_num = int(all_cpu-1)
    else:
        real_work_num = 1
    return real_work_num


# 任务结束写日志
# 设置回调函数
def setcallback_write_log(x):
    log_path = x[2]
    with open(log_path, 'a+') as f:
        line = str(x[0]) + str(x[1]) + "\n"
        f.write(line)
    with open(log_path.replace("todo0.log.txt","todo1.txt"), 'a+') as f:
        line = str(x[0]) + "\n"
        f.write(line)

# def quality_is_ok(file_path):
#     """检测文件degree角度是否大于30，大于30或者文件为空则返回False，否则为True，要保证文件存在

#     Args:
#         file_path ([type]): [description]

#     Returns:
#         [type]: [description]
#     """
#     with open(file_path, 'r')as f:
#         file_data = f.read()
#     # print(file_data)
#     pattern = r'Angle.*?Degree'  # 模式字符串
#     string = file_data  #要匹配的字符串
#     match = re.findall(pattern,string)
#     # print(match)
#     p = r'\d+\.?\d+'
#     nums = re.findall(p,str(match))
#     if len(nums) == 0:
#         result=False
#     else:
#         result = True
#         for num in nums:
#             if float(num) > 30:
#                 result_n = False
#             else:
#                 result_n = True
#             result = result and result_n
#     return result

# ------局部具体任务-----------------------------------------
# 单个任务控制者
def manager_one_job(data_root, job, save_path):
    # grid_level=str(grid_level)
    # num_quality = run_one_quality(job,save_path)
    # if "4" in grid_level:  # 质量控制
    #     num_quality = run_one_quality(job,save_path)
    # if "0" in grid_level:  # 预处理
    result = run_one_0(data_root, job,save_path)
    # if "1" in grid_level or "2" in grid_level or "3" in grid_level: # 计算
    #     result = run_one_123(data_root, job, app_root, save_path,runspace_root,grid_level)
    
    
    # "/home/zzr/project/linff/linff-main/deploy/space/log"
    log_path = os.path.join(save_path, "todo0.log.txt")
    result = [job, result, log_path]

    return result  # ,log_path
    # return job,result

# def run_one_quality(job, save_path):
#     """检测该任务质量是否通过
#     3质量3通过
#     2质量2通过，且没有质量3
#     1质量1通过且没有质量23
#     4没有质量文件,
#     5没有质量文件，但是有pre文件
#     False质量未通过删除文件夹

#     Returns:
#         [type]: [description]
#     """
#     job_name_list = job.split(".")
#     num = int(job_name_list[2])
#     num_level_1 = num//1000
#     num_level_2 = (num//100) % 10
#     this_dir = "num_{}{}00_{}{}99".format(
#         num_level_1, num_level_2, num_level_1, num_level_2)
#     job_save_path = os.path.join(save_path, this_dir, job)
#     p_jq3 = os.path.join(job_save_path,"NLFFFquality3.log")
#     p_jq2 = os.path.join(job_save_path,"NLFFFquality2.log")
#     p_jq1 = os.path.join(job_save_path,"NLFFFquality1.log")

#     if os.path.exists(p_jq3):
#         q3=quality_is_ok(p_jq3)
#         if q3:
#             result=3
#         else:
#             shutil.rmtree(job_save_path)
#             result=False
#     elif os.path.exists(p_jq2):
#         q2=quality_is_ok(p_jq2)
#         if q2:
#             result=2
#         else:
#             shutil.rmtree(job_save_path)
#             result=False
#     elif os.path.exists(p_jq1):
#         q1=quality_is_ok(p_jq1)
#         if q1:
#             result=1
#         else:
#             shutil.rmtree(job_save_path)
#             result=False
#     # elif check_have_pre(job, save_path):
#     #     result = 5
#     else:
#         shutil.rmtree(job_save_path)
#         result = 4 
#     return result

# # def check_have_pre(job,save_path):
    
#     return True

def run_one_0(data_root, job,save_path):
    print("0")
    # num_quality = run_one_quality(job,save_path)
    # if num_quality is False or num_quality == 4:
    job_save_path = make_job_workspace_dir(job,save_path)
    run_one_pre_result = run_one_pre(data_root, job, job_save_path)
    # job_save_path_remote = make_job_workspace_dir(job, save_path)
    # result = run_move_to_remote(run_one_pre_result,job_save_path_remote)
    # else:
    # result = True
    return run_one_pre_result


# def run_one_123(data_root, job, app_root, save_path,runspace_root,grid_level):
#     print("123") # 跑123 前面必须有跑过0
#     remote_job_save_path = make_job_workspace_dir(job, save_path)
#     remote_done_flag_file = os.path.join(remote_job_save_path, "done.txt")
#     if os.path.exists(remote_done_flag_file):
#         result = True
#     else:
#         job_save_path = make_job_workspace_dir(job, runspace_root)
#         com_c_path = os.path.join(app_root, "multigrid.sh")
#         com_c_exist = os.path.isfile(com_c_path)
#         if  com_c_exist:
#             result = run_one_comp(app_path=com_c_path, work_path=job_save_path,grid_level=grid_level)
#         else:
#             result = False
#             error_msg = "{} :运行脚本不存在或运行失败\n".format(job)
#             print(error_msg)
#             err_path = os.path.join(save_path, "run_err_log.txt")
#             with open(err_path, "a+") as f:
#                 f.write("{}:{}".format(job, error_msg))
#         if result:
#             result = run_one_clean(job_save_path)
#             if result:
#                 result = run_move_to_remote(job_save_path,remote_job_save_path)


def run_move_to_remote(from_dir,to_dir):
    # shutil.move("oldpos","newpos")
    try:
        if os.path.normcase(from_dir) == os.path.normcase(to_dir):
            result = True
        else:
            shutil.rmtree(to_dir)
            shutil.move(from_dir,to_dir)
            result = True
    except BaseException as e:
        print(e)
        result = False
    return result


# 根据任务文件生成文件夹和pre预处理
def run_one_pre(data_root, job, job_save_path):
    p_raw_path = os.path.join(data_root, job+".Bp.fits")
    t_raw_path = os.path.join(data_root, job+".Bt.fits")
    r_raw_path = os.path.join(data_root, job+".Br.fits")
    # print(p_raw_path)
    # print(os.path.isfile(p_raw_path))
    if os.path.isfile(p_raw_path) and os.path.isfile(t_raw_path) and os.path.isfile(r_raw_path):

        # job_save_path=make_job_workspace_dir(job,save_path)
        # print(job_save_path)
        try:
            pre_worker = PrepareWorker()
            pre_result = pre_worker.prepare_from_fits_Bprt(
                p_raw_path, t_raw_path, r_raw_path, job_save_path)
            result = pre_result
        except BaseException as e:
            error_msg = "{}预处理失败：{}".format(job, e)
            print(error_msg)
            err_path = os.path.join(job_save_path, "todo0.err.log.txt")
            with open(err_path, "a+") as f:
                f.write(error_msg)
            result = False
    return result


def make_job_workspace_dir(job, save_path):
    job_name_list = job.split(".")
    num = int(job_name_list[2])
    num_level_1 = num//1000
    num_level_2 = (num//100) % 10
    this_dir = "num_{}{}00_{}{}99".format(
        num_level_1, num_level_2, num_level_1, num_level_2)
    job_save_path = os.path.join(save_path, this_dir, job)
    if not os.path.exists(job_save_path):
        try:
            run_result = os.makedirs(job_save_path)
            if run_result is None:
                result = job_save_path
        except BaseException as e:
            error_msg = "{} ERROR: {}".format(job, e)
            print(error_msg)
            err_path = os.path.join(save_path, "todo0.err.log.txt")
            with open(err_path, "a+") as f:
                f.write("{}:{}".format(job, error_msg))
            result = False
    else:
        result = job_save_path
    return result

# 进入预处理执行主程序


# def run_one_comp(app_path, work_path, grid_level):
#     # app_path=app_path  # "/home/zzr/project/linff/linff-main/clinff/multigrid.sh"
#     # work_path=work_path #"/home/zzr/project/linff/temp/py2/a6"
#     # cmd_str="{} {}".format(app_path,work_path)
#     # # cmd_str="sssssss"
#     # cmd_tmp_log = subprocess.run([app_path,work_path],stdout=PIPE,stderr=STDOUT,) # https://blog.csdn.net/daduryi/article/details/81299999
#     # print(cmd_tmp_log)
#     # app_path="ls"
#     run_result = False
#     print("start run {}".format(work_path))
#     log_path = os.path.join(work_path, "run.log")
#     with open(log_path, 'w') as f:
#         # https://blog.csdn.net/daduryi/article/details/81299999
#         cmd_proc = subprocess.run(
#             [app_path, work_path,grid_level], stdout=f, stderr=subprocess.STDOUT)
#         # f.write(cmd_tmp_log.read()) # https://zhuanlan.zhihu.com/p/117495961
#         if cmd_proc.returncode == 0:
#             run_result = True
#     print("end run {}".format(work_path))
#     return run_result


# # 删除其他文件
# def run_one_clean(work_path):
#     return True  # 先不删除
#     remove_datalist = ["B0.bin"]
#     try:
#         for rm_file in remove_datalist:
#             rm_path = os.path.join(work_path, rm_file)
#             if os.path.isfile(rm_path):
#                 os.remove(rm_path)
#         result = True
#     except BaseException as e:
#         print("{} ERROR {} ".format(work_path, e))
#         result = False
#     return result


# -------主运行部分------------------------------------------
if __name__ == "__main__":
    main_control()
